"""
This package contains code that was "generated" via metaprogramming.

This happens, because Python is not flexible enough to do most tasks
common in typed functional programming.

Policy:

1. We store implementations in regular ``.py`` files.
2. We store generated type annotations in ``.pyi`` files.
3. We re-export these functions into regular modules as public values.

Please, do not touch this code unless you know what you are doing.
"""



from abc import abstractmethod
from typing import Generic, TypeVar

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')

_UnwrappableType = TypeVar('_UnwrappableType', bound='Unwrappable')


class Unwrappable(Generic[_FirstType, _SecondType]):
    """
    Represents containers that can unwrap and return its wrapped value.

    There are no aliases or ``UnwrappableN`` for ``Unwrappable`` interface.
    Because it always uses two and just two types.

    Not all types can be ``Unwrappable`` because we do require
    to raise ``UnwrapFailedError`` if unwrap is not possible.
    """

    __slots__ = ()

    @abstractmethod
    def unwrap(self: _UnwrappableType) -> _FirstType:
        """
        Custom magic method to unwrap inner value from container.

        Should be redefined for ones that actually have values.
        And for ones that raise an exception for no values.

        .. note::
            As a part of the contract, failed ``unwrap`` calls
            must raise :class:`returns.primitives.exceptions.UnwrapFailedError`
            exception.

        This method is the opposite of :meth:`~Unwrapable.failure`.
        """

    @abstractmethod
    def failure(self: _UnwrappableType) -> _SecondType:
        """
        Custom magic method to unwrap inner value from the failed container.

        .. note::
            As a part of the contract, failed ``failure`` calls
            must raise :class:`returns.primitives.exceptions.UnwrapFailedError`
            exception.

        This method is the opposite of :meth:`~Unwrapable.unwrap`.
        """



from typing import TypeVar

from returns.interfaces.unwrappable import Unwrappable
from returns.pipeline import is_successful

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')


def unwrap_or_failure(
    container: Unwrappable[_FirstType, _SecondType],
) -> _FirstType | _SecondType:
    """
    Unwraps either successful or failed value.

    .. code:: python

      >>> from returns.io import IO, IOSuccess, IOFailure
      >>> from returns.methods import unwrap_or_failure

      >>> assert unwrap_or_failure(IOSuccess(1)) == IO(1)
      >>> assert unwrap_or_failure(IOFailure('a')) == IO('a')

    """
    if is_successful(container):
        return container.unwrap()
    return container.failure()



import pytest

from returns.io import IO, IOFailure, IOSuccess
from returns.primitives.exceptions import UnwrapFailedError


def test_ioresult_value_or():
    """Ensures that ``value_or`` works correctly."""
    assert IOSuccess(1).value_or(0) == IO(1)
    assert IOFailure(1).value_or(0) == IO(0)


def test_unwrap_iosuccess():
    """Ensures that unwrap works for IOSuccess container."""
    assert IOSuccess(5).unwrap() == IO(5)


def test_unwrap_iofailure():
    """Ensures that unwrap works for IOFailure container."""
    with pytest.raises(UnwrapFailedError):
        IOFailure(5).unwrap()


def test_unwrap_iofailure_with_exception():
    """Ensures that unwrap raises from the original exception."""
    expected_exception = ValueError('error')
    with pytest.raises(UnwrapFailedError) as excinfo:
        IOFailure(expected_exception).unwrap()

    assert 'ValueError: error' in str(
        excinfo.getrepr(),  # noqa: WPS441
    )


def test_failure_iosuccess():
    """Ensures that failure works for IOSuccess container."""
    with pytest.raises(UnwrapFailedError):
        IOSuccess(5).failure()


def test_failure_iofailure():
    """Ensures that failure works for IOFailure container."""
    assert IOFailure(5).failure() == IO(5)



from abc import abstractmethod
from collections.abc import Callable
from typing import Generic, Never, TypeVar

from returns.primitives.hkt import KindN

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_LashableType = TypeVar('_LashableType', bound='LashableN')


class LashableN(Generic[_FirstType, _SecondType, _ThirdType]):
    """
    Represents a "context" in which calculations can be executed.

    ``Rescueable`` allows you to bind together
    a series of calculations while maintaining
    the context of that specific container.

    In contrast to :class:`returns.interfaces.bindable.BinbdaleN`,
    works with the second type value.
    """

    __slots__ = ()

    @abstractmethod
    def lash(
        self: _LashableType,
        function: Callable[
            [_SecondType],
            KindN[_LashableType, _FirstType, _UpdatedType, _ThirdType],
        ],
    ) -> KindN[_LashableType, _FirstType, _UpdatedType, _ThirdType]:
        """
        Applies 'function' to the result of a previous calculation.

        And returns a new container.
        """


#: Type alias for kinds with two type arguments.
Lashable2 = LashableN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
Lashable3 = LashableN[_FirstType, _SecondType, _ThirdType]



import pickle  # noqa: S403
from typing import Any

from hypothesis import example, given
from hypothesis import strategies as st

from returns.primitives.container import BaseContainer


class _CustomClass:
    def __init__(self, inner_value: Any) -> None:
        self.inner_value = inner_value

    def __eq__(self, other: object) -> bool:
        return (
            type(other) is type(self)  # noqa: WPS516
            and self.inner_value == other.inner_value
        )

    def __hash__(self) -> int:
        return hash(self.inner_value)


@given(
    st.one_of(
        st.integers(),
        st.floats(allow_nan=False),
        st.text(),
        st.booleans(),
        st.lists(st.text()),
        st.dictionaries(st.text(), st.integers()),
        st.builds(_CustomClass, st.text()),
    ),
)
@example(None)
def test_pickle(container_value: Any):
    """Ensures custom pickle protocol works as expected."""
    container = BaseContainer(container_value)
    assert pickle.loads(pickle.dumps(container)) == container  # noqa: S301



"""
Represents the base interfaces for types that do fearless async operations.

This type means that ``Future`` cannot fail.
Don't use this type for async that can. Instead, use
:class:`returns.interfaces.specific.future_result.FutureResultBasedN` type.
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Awaitable, Callable, Generator
from typing import TYPE_CHECKING, Any, Generic, Never, TypeVar

from returns.interfaces.specific import io
from returns.primitives.hkt import KindN

if TYPE_CHECKING:
    from returns.future import Future  # noqa: WPS433

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_FutureLikeType = TypeVar('_FutureLikeType', bound='FutureLikeN')
_AsyncFutureType = TypeVar('_AsyncFutureType', bound='AwaitableFutureN')


class FutureLikeN(io.IOLikeN[_FirstType, _SecondType, _ThirdType]):
    """
    Base type for ones that does look like ``Future``.

    But at the time this is not a real ``Future`` and cannot be awaited.
    """

    __slots__ = ()

    @abstractmethod
    def bind_future(
        self: _FutureLikeType,
        function: Callable[[_FirstType], Future[_UpdatedType]],
    ) -> KindN[_FutureLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Allows to bind ``Future`` returning function over a container."""

    @abstractmethod
    def bind_async_future(
        self: _FutureLikeType,
        function: Callable[[_FirstType], Awaitable[Future[_UpdatedType]]],
    ) -> KindN[_FutureLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Allows to bind async ``Future`` returning function over container."""

    @abstractmethod
    def bind_async(
        self: _FutureLikeType,
        function: Callable[
            [_FirstType],
            Awaitable[
                KindN[_FutureLikeType, _UpdatedType, _SecondType, _ThirdType],
            ],
        ],
    ) -> KindN[_FutureLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Binds async function returning the same type of container."""

    @abstractmethod
    def bind_awaitable(
        self: _FutureLikeType,
        function: Callable[[_FirstType], Awaitable[_UpdatedType]],
    ) -> KindN[_FutureLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Allows to bind async function over container."""

    @classmethod
    @abstractmethod
    def from_future(
        cls: type[_FutureLikeType],
        inner_value: Future[_UpdatedType],
    ) -> KindN[_FutureLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Unit method to create new containers from successful ``Future``."""


#: Type alias for kinds with one type argument.
FutureLike1 = FutureLikeN[_FirstType, Never, Never]

#: Type alias for kinds with two type arguments.
FutureLike2 = FutureLikeN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
FutureLike3 = FutureLikeN[_FirstType, _SecondType, _ThirdType]


class AwaitableFutureN(Generic[_FirstType, _SecondType, _ThirdType]):
    """
    Type that provides the required API for ``Future`` to be async.

    Should not be used directly. Use ``FutureBasedN`` instead.
    """

    __slots__ = ()

    @abstractmethod
    def __await__(
        self: _AsyncFutureType,
    ) -> Generator[
        Any,
        Any,
        io.IOLikeN[_FirstType, _SecondType, _ThirdType],
    ]:
        """Magic method to allow ``await`` expression."""

    @abstractmethod
    async def awaitable(
        self: _AsyncFutureType,
    ) -> io.IOLikeN[_FirstType, _SecondType, _ThirdType]:
        """Underling logic under ``await`` expression."""


#: Type alias for kinds with one type argument.
AsyncFuture1 = AwaitableFutureN[_FirstType, Never, Never]

#: Type alias for kinds with two type arguments.
AsyncFuture2 = AwaitableFutureN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
AsyncFuture3 = AwaitableFutureN[_FirstType, _SecondType, _ThirdType]


class FutureBasedN(
    FutureLikeN[_FirstType, _SecondType, _ThirdType],
    AwaitableFutureN[_FirstType, _SecondType, _ThirdType],
):
    """
    Base type for real ``Future`` objects.

    They can be awaited.
    """

    __slots__ = ()


#: Type alias for kinds with one type argument.
FutureBased1 = FutureBasedN[_FirstType, Never, Never]

#: Type alias for kinds with two type arguments.
FutureBased2 = FutureBasedN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
FutureBased3 = FutureBasedN[_FirstType, _SecondType, _ThirdType]



from returns.maybe import Nothing, Some


def test_some_is_true() -> None:
    """Ensures that ``Something(...)`` is ``True`` when treated as a boolean."""
    assert bool(Some(123))
    assert bool(Some('abc'))


def test_nothing_is_false() -> None:
    """Ensures that ``Nothing`` is ``False`` when treated as a boolean."""
    assert not bool(Nothing)


def test_some_none_is_true() -> None:
    """
    Ensures that ``Something(None)`` is ``True`` when treated as a boolean.

    See <https://github.com/dry-python/returns/issues/2177> for the discussion
    of this design choice.
    """
    assert bool(Some(None))
    assert bool(Some(Nothing))



def assert_equal(
    first,
    second,
    *,
    deps=None,
    backend: str = 'asyncio',
) -> None:
    """
    Custom ``assert`` function to compare two any containers.

    The important note here is that
    this ``assert`` should probably used in tests.
    Not real application code.

    It will call all ``Reader`` based containers and ``await``
    all ``Future`` based ones.

    It also works recursively.
    For example, ``ReaderFutureResult`` will be called and then awaited.

    You can specify different dependencies to call your containers.
    And different backends to ``await`` then using ``anyio``.

    By the way, ``anyio`` should be installed separately.
    """
    assert _convert(
        first,
        deps=deps,
        backend=backend,
    ) == _convert(
        second,
        deps=deps,
        backend=backend,
    ), f'{first} == {second}'


def _convert(container, *, deps, backend: str):
    from returns.interfaces.specific import future, reader  # noqa: PLC0415

    if isinstance(container, future.AwaitableFutureN):
        import anyio  # noqa: PLC0415

        return _convert(
            anyio.run(container.awaitable, backend=backend),
            deps=deps,
            backend=backend,
        )
    if isinstance(container, reader.Contextable):
        return _convert(
            container(deps),
            deps=deps,
            backend=backend,
        )
    return container



from collections.abc import Sequence

import pytest

from returns.context import (
    Reader,
    ReaderFutureResult,
    ReaderIOResult,
    ReaderResult,
)
from returns.contrib.pytest import ReturnsAsserts
from returns.future import Future, FutureResult
from returns.io import IO, IOResult
from returns.maybe import Maybe
from returns.primitives.asserts import assert_equal
from returns.primitives.container import BaseContainer
from returns.result import Result

_containers: Sequence[BaseContainer] = (
    Result.from_failure(1),
    Result.from_value(1),
    IO(1),
    IOResult.from_failure(1),
    IOResult.from_value(1),
    Maybe.from_value(1),
    Maybe.from_value(None),
    Maybe.from_optional(None),
    Future.from_value(1),
    FutureResult.from_value(1),
    FutureResult.from_failure(1),
    Reader.from_value(1),
    ReaderResult.from_value(1),
    ReaderResult.from_failure(1),
    ReaderIOResult.from_value(1),
    ReaderIOResult.from_failure(1),
    ReaderFutureResult.from_value(1),
    ReaderFutureResult.from_failure(1),
)


@pytest.mark.parametrize('container', _containers)
def test_assert_equal(container, anyio_backend_name: str):
    """Ensure that containers can be equal."""
    assert_equal(container, container, backend=anyio_backend_name)


@pytest.mark.parametrize('container', _containers)
def test_assert_equal_plugin(
    container,
    anyio_backend_name: str,
    returns: ReturnsAsserts,
):
    """Ensure that containers can be equal."""
    returns.assert_equal(container, container, backend=anyio_backend_name)


@pytest.mark.parametrize('container', _containers)
def test_assert_equal_not(container, anyio_backend_name: str):
    """Ensure that containers can be not equal."""
    with pytest.raises(AssertionError):
        assert_equal(
            container,
            container.from_value(2),
            backend=anyio_backend_name,
        )


@pytest.mark.parametrize('container', _containers)
def test_assert_equal_not_plugin(
    container,
    anyio_backend_name: str,
    returns: ReturnsAsserts,
):
    """Ensure that containers can be not equal."""
    with pytest.raises(AssertionError):
        returns.assert_equal(
            container,
            container.from_value(2),
            backend=anyio_backend_name,
        )



from returns.functions import compose


def _first(argument: int) -> str:
    return str(argument)


def _second(argument: str) -> bool:
    return bool(argument)


def test_function_composition():
    """Ensures that functions can be composed and return type is correct."""
    second_after_first = compose(_first, _second)

    assert second_after_first(1) is True
    assert second_after_first(0) is True



from collections.abc import Awaitable
from typing import Any

import pytest

from returns.future import Future, FutureResult
from returns.io import IO


@pytest.mark.anyio
async def test_inner_value(subtests):
    """Ensure that coroutine correct value is preserved for all units."""
    containers: list[Awaitable[Any]] = [
        # We have to define these values inside the test, because
        # otherwise `anyio` will `await` reused coroutines.
        # And they have to be fresh. That's why we use subtests for it.
        Future.from_value(1),
        Future.from_io(IO(1)),
        Future.from_future_result(FutureResult.from_value(1)),
        Future.from_future_result(FutureResult.from_failure(1)),
    ]
    for container in containers:
        with subtests.test(container=container):
            assert isinstance(await container, IO)



from typing import Generic, TypeVar

from hypothesis import strategies as st

from returns.contrib.hypothesis.type_resolver import (
    look_up_strategy,
    strategies_for_types,
)

_ValueType = TypeVar('_ValueType')


class _Wrapper1(Generic[_ValueType]):
    _inner_value: _ValueType


class _Wrapper2(Generic[_ValueType]):
    _inner_value: _ValueType


def test_types_without_strategies() -> None:  # noqa: WPS210
    """Check that it temporarily resolves a type that has no strategy."""
    strategy_before1 = look_up_strategy(_Wrapper1)
    strategy_before2 = look_up_strategy(_Wrapper2)

    with strategies_for_types({
        _Wrapper1: st.builds(_Wrapper1, st.integers()),
        _Wrapper2: st.builds(_Wrapper2, st.text()),
    }):
        strategy_inside1 = look_up_strategy(_Wrapper1)
        strategy_inside2 = look_up_strategy(_Wrapper2)

    strategy_after1 = look_up_strategy(_Wrapper1)
    strategy_after2 = look_up_strategy(_Wrapper2)

    assert strategy_before1 is None
    assert strategy_before2 is None
    assert str(strategy_inside1) == 'builds(_Wrapper1, integers())'
    assert str(strategy_inside2) == 'builds(_Wrapper2, text())'
    assert strategy_after1 is None
    assert strategy_after2 is None


def test_type_with_strategy() -> None:
    """Check that it restores the original strategy."""
    with strategies_for_types({_Wrapper1: st.builds(_Wrapper1, st.integers())}):
        strategy_before = look_up_strategy(_Wrapper1)

        with strategies_for_types({_Wrapper1: st.builds(_Wrapper1, st.text())}):
            strategy_inside = look_up_strategy(_Wrapper1)

        strategy_after = look_up_strategy(_Wrapper1)

    assert str(strategy_before) == 'builds(_Wrapper1, integers())'
    assert str(strategy_inside) == 'builds(_Wrapper1, text())'
    assert str(strategy_after) == 'builds(_Wrapper1, integers())'



from returns.context import RequiresContext
from returns.context import RequiresContextResult as RCR  # noqa: N817
from returns.result import Failure, Result, Success


def test_bind():
    """Ensures that bind works."""

    def factory(inner_value: int) -> RCR[float, str, int]:
        if inner_value > 0:
            return RCR(lambda deps: Success(inner_value / deps))
        return RCR.from_failure(str(inner_value))

    input_value = 5
    bound: RCR[int, str, int] = RCR.from_value(input_value)
    assert bound.bind(factory)(2) == factory(input_value)(2)
    assert bound.bind(factory)(2) == Success(2.5)

    assert (
        RCR.from_value(0).bind(
            factory,
        )(2)
        == factory(0)(2)
        == Failure('0')
    )


def test_bind_regular_result():
    """Ensures that regular ``Result`` can be bound."""

    def factory(inner_value: int) -> Result[int, str]:
        if inner_value > 0:
            return Success(inner_value + 1)
        return Failure('nope')

    first: RCR[int, str, int] = RCR.from_value(1)
    third: RCR[int, str, int] = RCR.from_failure('a')

    assert first.bind_result(factory)(RCR.no_args) == Success(2)
    assert RCR.from_value(0).bind_result(
        factory,
    )(RCR.no_args) == Failure('nope')
    assert third.bind_result(factory)(RCR.no_args) == Failure('a')


def test_bind_regular_context():
    """Ensures that regular ``RequiresContext`` can be bound."""

    def factory(inner_value: int) -> RequiresContext[float, int]:
        return RequiresContext(lambda deps: inner_value / deps)

    first: RCR[int, str, int] = RCR.from_value(1)
    third: RCR[int, str, int] = RCR.from_failure('a')

    assert first.bind_context(factory)(2) == Success(0.5)
    assert RCR.from_value(2).bind_context(
        factory,
    )(1) == Success(2.0)
    assert third.bind_context(factory)(1) == Failure('a')


def test_lash_success():
    """Ensures that lash works for Success container."""

    def factory(inner_value) -> RCR[int, str, int]:
        return RCR.from_value(inner_value * 2)

    assert RCR.from_value(5).lash(
        factory,
    )(0) == RCR.from_value(5)(0)
    assert RCR.from_failure(5).lash(
        factory,
    )(0) == RCR.from_value(10)(0)


def test_lash_failure():
    """Ensures that lash works for Failure container."""

    def factory(inner_value) -> RCR[int, str, int]:
        return RCR.from_failure(inner_value * 2)

    assert RCR.from_value(5).lash(
        factory,
    )(0) == RCR.from_value(5)(0)
    assert RCR.from_failure(5).lash(
        factory,
    )(0) == RCR.from_failure(10)(0)



from collections.abc import Callable, Sequence
from typing import Any, TypeVar

import pytest
from hypothesis import given
from hypothesis import strategies as st

from returns.context import (
    Reader,
    RequiresContext,
    RequiresContextFutureResult,
    RequiresContextFutureResultE,
    RequiresContextIOResult,
    RequiresContextIOResultE,
    RequiresContextResult,
    RequiresContextResultE,
)
from returns.contrib.hypothesis.laws import (
    Settings,
    _types_to_strategies,  # noqa: PLC2701
    default_settings,
)
from returns.contrib.hypothesis.type_resolver import (
    StrategyFactory,
    apply_strategy,
    look_up_strategy,
)
from returns.future import Future, FutureResult
from returns.interfaces.applicative import ApplicativeN
from returns.io import IO, IOResult, IOResultE
from returns.maybe import Maybe
from returns.pipeline import is_successful
from returns.primitives.laws import Lawful
from returns.result import Result, ResultE
from test_hypothesis.test_laws import test_custom_type_applicative

_all_containers: Sequence[type[Lawful]] = (
    Maybe,
    Result,
    IO,
    IOResult,
    Future,
    FutureResult,
    RequiresContext,
    RequiresContextResult,
    RequiresContextIOResult,
    RequiresContextFutureResult,
    # Aliases:
    ResultE,
    IOResultE,
    Reader,
    RequiresContextResultE,
    RequiresContextIOResultE,
    RequiresContextFutureResultE,
)


@pytest.mark.filterwarnings('ignore:.*')
@pytest.mark.parametrize('container_type', _all_containers)
def test_all_containers_resolves(container_type: type[Lawful]) -> None:
    """Ensures all containers do resolve."""
    assert st.from_type(container_type).example() is not None


@given(
    st.from_type(ResultE).filter(
        lambda container: not is_successful(container),
    ),
)
def test_result_error_alias_resolves(thing: ResultE[Any]) -> None:
    """Ensures that type aliases are resolved correctly."""
    assert isinstance(thing.failure(), Exception)


CustomResult = Result[int, str]


@given(st.from_type(CustomResult))
def test_custom_result_error_types_resolve(thing: CustomResult) -> None:
    """Ensures that type aliases are resolved correctly."""
    if is_successful(thing):
        assert isinstance(thing.unwrap(), int)
    else:
        assert isinstance(thing.failure(), str)


@given(
    st.from_type(RequiresContextResultE).filter(
        lambda container: (
            not is_successful(
                container(RequiresContextResultE.no_args),
            )
        ),
    ),
)
def test_reader_result_error_alias_resolves(
    thing: RequiresContextResultE,
) -> None:
    """Ensures that type aliases are resolved correctly."""
    real_result = thing(RequiresContextResultE.no_args)
    assert isinstance(real_result.failure(), Exception)


CustomReaderResult = RequiresContextResult[int, str, bool]


@given(st.from_type(CustomReaderResult))
def test_custom_readerresult_types_resolve(
    thing: CustomReaderResult,
) -> None:
    """Ensures that type aliases are resolved correctly."""
    real_result = thing(RequiresContextResultE.no_args)
    if is_successful(real_result):
        assert isinstance(real_result.unwrap(), int)
    else:
        assert isinstance(real_result.failure(), str)


def test_merge_settings() -> None:
    """Check that each part of the settings can be overridden by users."""
    settings1 = Settings(
        settings_kwargs={'a': 1, 'b': 2},
        use_init=False,
        container_strategy=st.integers(),
        type_strategies={int: st.integers(max_value=10), str: st.text('abc')},
    )
    settings2 = Settings(
        settings_kwargs={'a': 1, 'c': 3},
        use_init=False,
        container_strategy=st.integers(max_value=20),
        type_strategies={int: st.integers(max_value=30), bool: st.booleans()},
    )

    result = settings1 | settings2

    assert result == Settings(
        settings_kwargs={'a': 1, 'b': 2, 'c': 3},
        use_init=False,
        container_strategy=st.integers(max_value=20),
        type_strategies={
            int: st.integers(max_value=30),
            bool: st.booleans(),
            str: st.text('abc'),
        },
    )


def test_merge_use_init() -> None:
    """Check that `use_init` can be set to `True` by users.

    Note: They can't set a `True` to `False`, since we use `|` to merge.
    However, the default value is `False`, so this should not be a problem.
    """
    settings1 = Settings(
        settings_kwargs={},
        use_init=False,
        container_strategy=None,
        type_strategies={},
    )
    settings2 = Settings(
        settings_kwargs={},
        use_init=True,
        container_strategy=None,
        type_strategies={},
    )

    result = settings1 | settings2

    assert result == Settings(
        settings_kwargs={},
        use_init=True,
        container_strategy=None,
        type_strategies={},
    )


_ValueType = TypeVar('_ValueType')


def test_types_to_strategies_default() -> None:  # noqa: WPS210
    """Check the default strategies for types."""
    container_type = test_custom_type_applicative._Wrapper  # noqa: SLF001
    # NOTE: There is a type error because `Callable` is a
    # special form, not a type.
    callable_type: type[object] = Callable  # type: ignore[assignment]

    result = _types_to_strategies(
        container_type,
        default_settings(container_type),
    )

    wrapper_strategy = (
        'builds(from_value, shared(sampled_from([NoneType,'
        ' bool, int, float, str,'
        " bytes]), key='typevar=~_FirstType').flatmap(from_type))"
    )
    assert (
        _strategy_string(result[container_type], container_type)
        == wrapper_strategy
    )
    assert _strategy_strings(
        [result[interface] for interface in container_type.laws()],
        container_type,
    ) == [
        wrapper_strategy,
        wrapper_strategy,
    ]
    assert (
        _strategy_string(result[callable_type], Callable[[int, str], bool])
        == 'functions(like=lambda *args, **kwargs: None,'
        ' returns=booleans(), pure=True)'
    )
    assert (
        _strategy_string(result[callable_type], Callable[[], None])
        == 'functions(like=lambda: None, returns=none(), pure=True)'
    )
    assert (
        _strategy_string(result[TypeVar], _ValueType)
        == 'shared(sampled_from([NoneType, bool,'
        ' int, float, str, bytes]),'
        " key='typevar=~_ValueType').flatmap(from_type).filter(lambda"
        ' inner: inner == inner)'
    )


def test_types_to_strategies_overrides() -> None:  # noqa: WPS210
    """Check that we allow the user to override all strategies."""
    container_type = test_custom_type_applicative._Wrapper  # noqa: SLF001
    # NOTE: There is a type error because `Callable` is a
    # special form, not a type.
    callable_type: type[object] = Callable  # type: ignore[assignment]

    result = _types_to_strategies(
        container_type,
        Settings(
            settings_kwargs={},
            use_init=False,
            container_strategy=st.builds(container_type, st.integers()),
            type_strategies={
                TypeVar: st.text(),
                callable_type: st.functions(returns=st.booleans()),
                # This strategy does not get used, because we use
                # the given `container_strategy` for all interfaces of the
                # container type.
                ApplicativeN: st.tuples(st.integers()),
            },
        ),
    )

    wrapper_strategy = 'builds(_Wrapper, integers())'
    assert (
        _strategy_string(result[container_type], container_type)
        == wrapper_strategy
    )
    assert _strategy_strings(
        [result[interface] for interface in container_type.laws()],
        container_type,
    ) == [
        wrapper_strategy,
        wrapper_strategy,
    ]
    assert (
        _strategy_string(result[callable_type], Callable[[int, str], bool])
        == 'functions(returns=booleans())'
    )
    assert (
        _strategy_string(result[callable_type], Callable[[], None])
        == 'functions(returns=booleans())'
    )
    assert _strategy_string(result[TypeVar], _ValueType) == 'text()'


def _interface_factories(type_: type[Lawful]) -> list[StrategyFactory | None]:
    return [look_up_strategy(interface) for interface in type_.laws()]


def _strategy_strings(
    strategy_factories: Sequence[StrategyFactory | None], type_: type[object]
) -> list[str]:
    return [
        _strategy_string(strategy_factory, type_)
        for strategy_factory in strategy_factories
    ]


def _strategy_string(
    strategy_factory: StrategyFactory | None, type_: Any
) -> str:
    """Return an easily testable string representation."""
    return (
        str(None)
        if strategy_factory is None
        else str(apply_strategy(strategy_factory, type_))
    )



from abc import ABC
from collections.abc import Callable, Generator, Iterator
from functools import wraps
from inspect import FrameInfo
from typing import (
    TYPE_CHECKING,
    Any,
    Never,
    TypeAlias,
    TypeVar,
    final,
    overload,
)

from typing_extensions import ParamSpec

from returns.interfaces.specific import result
from returns.primitives.container import BaseContainer, container_equality
from returns.primitives.exceptions import UnwrapFailedError
from returns.primitives.hkt import Kind2, SupportsKind2

# Definitions:
_ValueType_co = TypeVar('_ValueType_co', covariant=True)
_NewValueType = TypeVar('_NewValueType')
_ErrorType_co = TypeVar('_ErrorType_co', covariant=True)
_NewErrorType = TypeVar('_NewErrorType')

_FirstType = TypeVar('_FirstType')
_FuncParams = ParamSpec('_FuncParams')


class Result(  # type: ignore[type-var]
    BaseContainer,
    SupportsKind2['Result', _ValueType_co, _ErrorType_co],
    result.ResultBased2[_ValueType_co, _ErrorType_co],
    ABC,
):
    """
    Base class for :class:`~Failure` and :class:`~Success`.

    :class:`~Result` does not have a public constructor.
    Use :func:`~Success` and :func:`~Failure` to construct the needed values.

    See also:
        - https://bit.ly/361qQhi
        - https://hackernoon.com/the-throw-keyword-was-a-mistake-l9e532di

    """

    __slots__ = ('_trace',)
    __match_args__ = ('_inner_value',)

    _inner_value: _ValueType_co | _ErrorType_co
    _trace: list[FrameInfo] | None

    #: Typesafe equality comparison with other `Result` objects.
    equals = container_equality

    @property
    def trace(self) -> list[FrameInfo] | None:
        """Returns a list with stack trace when :func:`~Failure` was called."""
        return self._trace

    def swap(self) -> 'Result[_ErrorType_co, _ValueType_co]':
        """
        Swaps value and error types.

        So, values become errors and errors become values.
        It is useful when you have to work with errors a lot.
        And since we have a lot of ``.bind_`` related methods
        and only a single ``.lash`` - it is easier to work with values.

        .. code:: python

          >>> from returns.result import Success, Failure
          >>> assert Success(1).swap() == Failure(1)
          >>> assert Failure(1).swap() == Success(1)

        """

    def map(
        self,
        function: Callable[[_ValueType_co], _NewValueType],
    ) -> 'Result[_NewValueType, _ErrorType_co]':
        """
        Composes successful container with a pure function.

        .. code:: python

          >>> from returns.result import Failure, Success

          >>> def mappable(string: str) -> str:
          ...      return string + 'b'

          >>> assert Success('a').map(mappable) == Success('ab')
          >>> assert Failure('a').map(mappable) == Failure('a')

        """

    def apply(
        self,
        container: Kind2[
            'Result',
            Callable[[_ValueType_co], _NewValueType],
            _ErrorType_co,
        ],
    ) -> 'Result[_NewValueType, _ErrorType_co]':
        """
        Calls a wrapped function in a container on this container.

        .. code:: python

          >>> from returns.result import Failure, Success

          >>> def appliable(string: str) -> str:
          ...      return string + 'b'

          >>> assert Success('a').apply(Success(appliable)) == Success('ab')
          >>> assert Failure('a').apply(Success(appliable)) == Failure('a')

          >>> assert Success('a').apply(Failure(1)) == Failure(1)
          >>> assert Failure(1).apply(Failure(2)) == Failure(1)

        """

    def bind(
        self,
        function: Callable[
            [_ValueType_co],
            Kind2['Result', _NewValueType, _ErrorType_co],
        ],
    ) -> 'Result[_NewValueType, _ErrorType_co]':
        """
        Composes successful container with a function that returns a container.

        .. code:: python

          >>> from returns.result import Result, Success, Failure

          >>> def bindable(arg: str) -> Result[str, str]:
          ...      if len(arg) > 1:
          ...          return Success(arg + 'b')
          ...      return Failure(arg + 'c')

          >>> assert Success('aa').bind(bindable) == Success('aab')
          >>> assert Success('a').bind(bindable) == Failure('ac')
          >>> assert Failure('a').bind(bindable) == Failure('a')

        """

    #: Alias for `bind_result` method, it is the same as `bind` here.
    bind_result = bind

    def alt(
        self,
        function: Callable[[_ErrorType_co], _NewErrorType],
    ) -> 'Result[_ValueType_co, _NewErrorType]':
        """
        Composes failed container with a pure function to modify failure.

        .. code:: python

          >>> from returns.result import Failure, Success

          >>> def altable(arg: str) -> str:
          ...      return arg + 'b'

          >>> assert Success('a').alt(altable) == Success('a')
          >>> assert Failure('a').alt(altable) == Failure('ab')

        """

    def lash(
        self,
        function: Callable[
            [_ErrorType_co],
            Kind2['Result', _ValueType_co, _NewErrorType],
        ],
    ) -> 'Result[_ValueType_co, _NewErrorType]':
        """
        Composes failed container with a function that returns a container.

        .. code:: python

          >>> from returns.result import Result, Success, Failure

          >>> def lashable(arg: str) -> Result[str, str]:
          ...      if len(arg) > 1:
          ...          return Success(arg + 'b')
          ...      return Failure(arg + 'c')

          >>> assert Success('a').lash(lashable) == Success('a')
          >>> assert Failure('a').lash(lashable) == Failure('ac')
          >>> assert Failure('aa').lash(lashable) == Success('aab')

        """

    def __iter__(self) -> Iterator[_ValueType_co]:
        """API for :ref:`do-notation`."""
        yield self.unwrap()

    @classmethod
    def do(
        cls,
        expr: Generator[_NewValueType, None, None],
    ) -> 'Result[_NewValueType, _NewErrorType]':
        """
        Allows working with unwrapped values of containers in a safe way.

        .. code:: python

          >>> from returns.result import Result, Failure, Success

          >>> assert Result.do(
          ...     first + second
          ...     for first in Success(2)
          ...     for second in Success(3)
          ... ) == Success(5)

          >>> assert Result.do(
          ...     first + second
          ...     for first in Failure('a')
          ...     for second in Success(3)
          ... ) == Failure('a')

        See :ref:`do-notation` to learn more.
        This feature requires our :ref:`mypy plugin <mypy-plugins>`.

        """
        try:
            return Result.from_value(next(expr))
        except UnwrapFailedError as exc:
            return exc.halted_container  # type: ignore

    def value_or(
        self,
        default_value: _NewValueType,
    ) -> _ValueType_co | _NewValueType:
        """
        Get value or default value.

        .. code:: python

          >>> from returns.result import Failure, Success
          >>> assert Success(1).value_or(2) == 1
          >>> assert Failure(1).value_or(2) == 2

        """

    def unwrap(self) -> _ValueType_co:
        """
        Get value or raise exception.

        .. code:: pycon
          :force:

          >>> from returns.result import Failure, Success
          >>> assert Success(1).unwrap() == 1

          >>> Failure(1).unwrap()
          Traceback (most recent call last):
            ...
          returns.primitives.exceptions.UnwrapFailedError

        """

    def failure(self) -> _ErrorType_co:
        """
        Get failed value or raise exception.

        .. code:: pycon
          :force:

          >>> from returns.result import Failure, Success
          >>> assert Failure(1).failure() == 1

          >>> Success(1).failure()
          Traceback (most recent call last):
            ...
          returns.primitives.exceptions.UnwrapFailedError

        """

    @classmethod
    def from_value(
        cls,
        inner_value: _NewValueType,
    ) -> 'Result[_NewValueType, Any]':
        """
        One more value to create success unit values.

        It is useful as a united way to create a new value from any container.

        .. code:: python

          >>> from returns.result import Result, Success
          >>> assert Result.from_value(1) == Success(1)

        You can use this method or :func:`~Success`,
        choose the most convenient for you.

        """
        return Success(inner_value)

    @classmethod
    def from_failure(
        cls,
        inner_value: _NewErrorType,
    ) -> 'Result[Any, _NewErrorType]':
        """
        One more value to create failure unit values.

        It is useful as a united way to create a new value from any container.

        .. code:: python

          >>> from returns.result import Result, Failure
          >>> assert Result.from_failure(1) == Failure(1)

        You can use this method or :func:`~Failure`,
        choose the most convenient for you.

        """
        return Failure(inner_value)

    @classmethod
    def from_result(
        cls,
        inner_value: 'Result[_NewValueType, _NewErrorType]',
    ) -> 'Result[_NewValueType, _NewErrorType]':
        """
        Creates a new ``Result`` instance from existing ``Result`` instance.

        .. code:: python

          >>> from returns.result import Result, Failure, Success
          >>> assert Result.from_result(Success(1)) == Success(1)
          >>> assert Result.from_result(Failure(1)) == Failure(1)

        This is a part of
        :class:`returns.interfaces.specific.result.ResultBasedN` interface.
        """
        return inner_value


@final  # noqa: WPS338
class Failure(Result[Any, _ErrorType_co]):  # noqa: WPS338
    """
    Represents a calculation which has failed.

    It should contain an error code or message.
    """

    __slots__ = ()

    _inner_value: _ErrorType_co

    def __init__(self, inner_value: _ErrorType_co) -> None:
        """Failure constructor."""
        super().__init__(inner_value)
        object.__setattr__(self, '_trace', self._get_trace())

    if not TYPE_CHECKING:  # noqa: WPS604  # pragma: no branch

        def alt(self, function):
            """Composes failed container with a pure function to modify failure."""  # noqa: E501
            return Failure(function(self._inner_value))

        def map(self, function):
            """Does nothing for ``Failure``."""
            return self

        def bind(self, function):
            """Does nothing for ``Failure``."""
            return self

        #: Alias for `bind` method. Part of the `ResultBasedN` interface.
        bind_result = bind

        def lash(self, function):
            """Composes this container with a function returning container."""
            return function(self._inner_value)

        def apply(self, container):
            """Does nothing for ``Failure``."""
            return self

        def value_or(self, default_value):
            """Returns default value for failed container."""
            return default_value

    def swap(self):
        """Failures swap to :class:`Success`."""
        return Success(self._inner_value)

    def unwrap(self) -> Never:
        """Raises an exception, since it does not have a value inside."""
        if isinstance(self._inner_value, Exception):
            raise UnwrapFailedError(self) from self._inner_value
        raise UnwrapFailedError(self)

    def failure(self) -> _ErrorType_co:
        """Returns failed value."""
        return self._inner_value

    def _get_trace(self) -> list[FrameInfo] | None:
        """Method that will be monkey patched when trace is active."""


@final
class Success(Result[_ValueType_co, Any]):
    """
    Represents a calculation which has succeeded and contains the result.

    Contains the computation value.
    """

    __slots__ = ()

    _inner_value: _ValueType_co

    def __init__(self, inner_value: _ValueType_co) -> None:
        """Success constructor."""
        super().__init__(inner_value)

    if not TYPE_CHECKING:  # noqa: WPS604  # pragma: no branch

        def alt(self, function):
            """Does nothing for ``Success``."""
            return self

        def map(self, function):
            """Composes current container with a pure function."""
            return Success(function(self._inner_value))

        def bind(self, function):
            """Binds current container to a function that returns container."""
            return function(self._inner_value)

        #: Alias for `bind` method. Part of the `ResultBasedN` interface.
        bind_result = bind

        def lash(self, function):
            """Does nothing for ``Success``."""
            return self

        def apply(self, container):
            """Calls a wrapped function in a container on this container."""
            if isinstance(container, Success):
                return self.map(container.unwrap())
            return container

        def value_or(self, default_value):
            """Returns the value for successful container."""
            return self._inner_value

    def swap(self):
        """Successes swap to :class:`Failure`."""
        return Failure(self._inner_value)

    def unwrap(self) -> _ValueType_co:
        """Returns the unwrapped value from successful container."""
        return self._inner_value

    def failure(self) -> Never:
        """Raises an exception for successful container."""
        raise UnwrapFailedError(self)


# Aliases:

#: Alias for ``Result[_ValueType_co, Exception]``.
ResultE: TypeAlias = Result[_ValueType_co, Exception]


# Decorators:

_ExceptionType = TypeVar('_ExceptionType', bound=Exception)


@overload
def safe(
    function: Callable[_FuncParams, _ValueType_co],
    /,
) -> Callable[_FuncParams, ResultE[_ValueType_co]]: ...


@overload
def safe(
    exceptions: tuple[type[_ExceptionType], ...],
) -> Callable[
    [Callable[_FuncParams, _ValueType_co]],
    Callable[_FuncParams, Result[_ValueType_co, _ExceptionType]],
]: ...


def safe(  # noqa: WPS234
    exceptions: (
        Callable[_FuncParams, _ValueType_co] | tuple[type[_ExceptionType], ...]
    ),
) -> (
    Callable[_FuncParams, ResultE[_ValueType_co]]
    | Callable[
        [Callable[_FuncParams, _ValueType_co]],
        Callable[_FuncParams, Result[_ValueType_co, _ExceptionType]],
    ]
):
    """
    Decorator to convert exception-throwing function to ``Result`` container.

    Should be used with care, since it only catches ``Exception`` subclasses.
    It does not catch ``BaseException`` subclasses.

    If you need to mark ``async`` function as ``safe``,
    use :func:`returns.future.future_safe` instead.
    This decorator only works with sync functions. Example:

    .. code:: python

      >>> from returns.result import Failure, Success, safe

      >>> @safe
      ... def might_raise(arg: int) -> float:
      ...     return 1 / arg

      >>> assert might_raise(1) == Success(1.0)
      >>> assert isinstance(might_raise(0), Failure)

    You can also use it with explicit exception types as the first argument:

    .. code:: python

      >>> from returns.result import Failure, Success, safe

      >>> @safe(exceptions=(ZeroDivisionError,))
      ... def might_raise(arg: int) -> float:
      ...     return 1 / arg

      >>> assert might_raise(1) == Success(1.0)
      >>> assert isinstance(might_raise(0), Failure)

    In this case, only exceptions that are explicitly
    listed are going to be caught.

    Similar to :func:`returns.io.impure_safe`
    and :func:`returns.future.future_safe` decorators.
    """

    def factory(
        inner_function: Callable[_FuncParams, _ValueType_co],
        inner_exceptions: tuple[type[_ExceptionType], ...],
    ) -> Callable[_FuncParams, Result[_ValueType_co, _ExceptionType]]:
        @wraps(inner_function)
        def decorator(
            *args: _FuncParams.args,
            **kwargs: _FuncParams.kwargs,
        ) -> Result[_ValueType_co, _ExceptionType]:
            try:
                return Success(inner_function(*args, **kwargs))
            except inner_exceptions as exc:
                return Failure(exc)

        return decorator

    if isinstance(exceptions, tuple):
        return lambda function: factory(function, exceptions)
    return factory(
        exceptions,
        (Exception,),  # type: ignore[arg-type]
    )


def attempt(
    func: Callable[[_FirstType], _NewValueType],
) -> Callable[[_FirstType], Result[_NewValueType, _FirstType]]:
    """
    Decorator to convert exception-throwing function to ``Result`` container.

    It's very similar with :func:`returns.result.safe`, the difference is when
    an exception is raised it won't wrap that given exception into a Failure,
    it'll wrap the argument that lead to the exception.

    .. code:: python

        >>> import json
        >>> from typing import Dict, Any

        >>> from returns.result import Failure, Success, attempt

        >>> @attempt
        ... def parse_json(string: str) -> Dict[str, Any]:
        ...     return json.loads(string)

        >>> assert parse_json('{"key": "value"}') == Success({'key': 'value'})
        >>> assert parse_json('incorrect input') == Failure('incorrect input')

    """

    @wraps(func)
    def decorator(arg: _FirstType) -> Result[_NewValueType, _FirstType]:
        try:
            return Success(func(arg))
        except Exception:
            return Failure(arg)

    return decorator



from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import ClassVar, Never, TypeVar, final

from returns.interfaces import equable, failable, unwrappable
from returns.primitives.asserts import assert_equal
from returns.primitives.hkt import KindN
from returns.primitives.laws import (
    Law,
    Law2,
    Law3,
    Lawful,
    LawSpecDef,
    law_definition,
)

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_MaybeLikeType = TypeVar('_MaybeLikeType', bound='MaybeLikeN')

# New values:
_ValueType = TypeVar('_ValueType')

# Only used in laws:
_NewType1 = TypeVar('_NewType1')


@final
class _LawSpec(LawSpecDef):
    """
    Maybe laws.

    We need to be sure that
    ``.map``, ``.bind``, ``.bind_optional``, and ``.lash``
    works correctly for both successful and failed types.
    """

    __slots__ = ()

    @law_definition
    def map_short_circuit_law(
        container: 'MaybeLikeN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[[_FirstType], _NewType1],
    ) -> None:
        """Ensures that you cannot map from failures."""
        assert_equal(
            container.from_optional(None).map(function),
            container.from_optional(None),
        )

    @law_definition
    def bind_short_circuit_law(
        container: 'MaybeLikeN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[
            [_FirstType],
            KindN['MaybeLikeN', _NewType1, _SecondType, _ThirdType],
        ],
    ) -> None:
        """Ensures that you cannot bind from failures."""
        assert_equal(
            container.from_optional(None).bind(function),
            container.from_optional(None),
        )

    @law_definition
    def bind_optional_short_circuit_law(
        container: 'MaybeLikeN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[[_FirstType], _NewType1 | None],
    ) -> None:
        """Ensures that you cannot bind from failures."""
        assert_equal(
            container.from_optional(None).bind_optional(function),
            container.from_optional(None),
        )

    @law_definition
    def lash_short_circuit_law(
        raw_value: _FirstType,
        container: 'MaybeLikeN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[
            [_SecondType],
            KindN['MaybeLikeN', _FirstType, _NewType1, _ThirdType],
        ],
    ) -> None:
        """Ensures that you cannot lash a success."""
        assert_equal(
            container.from_value(raw_value).lash(function),
            container.from_value(raw_value),
        )

    @law_definition
    def unit_structure_law(
        container: 'MaybeLikeN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[[_FirstType], None],
    ) -> None:
        """Ensures ``None`` is treated specially."""
        assert_equal(
            container.bind_optional(function),
            container.from_optional(None),
        )


class MaybeLikeN(
    failable.SingleFailableN[_FirstType, _SecondType, _ThirdType],
    Lawful['MaybeLikeN[_FirstType, _SecondType, _ThirdType]'],
):
    """
    Type for values that do look like a ``Maybe``.

    For example, ``RequiresContextMaybe`` should be created from this interface.
    Cannot be unwrapped or compared.
    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (
        Law2(_LawSpec.map_short_circuit_law),
        Law2(_LawSpec.bind_short_circuit_law),
        Law2(_LawSpec.bind_optional_short_circuit_law),
        Law3(_LawSpec.lash_short_circuit_law),
        Law2(_LawSpec.unit_structure_law),
    )

    @abstractmethod
    def bind_optional(
        self: _MaybeLikeType,
        function: Callable[[_FirstType], _UpdatedType | None],
    ) -> KindN[_MaybeLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Binds a function that returns ``Optional`` values."""

    @classmethod
    @abstractmethod
    def from_optional(
        cls: type[_MaybeLikeType],
        inner_value: _ValueType | None,
    ) -> KindN[_MaybeLikeType, _ValueType, _SecondType, _ThirdType]:
        """Unit method to create containers from ``Optional`` value."""


#: Type alias for kinds with two type arguments.
MaybeLike2 = MaybeLikeN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
MaybeLike3 = MaybeLikeN[_FirstType, _SecondType, _ThirdType]


class MaybeBasedN(
    MaybeLikeN[_FirstType, _SecondType, _ThirdType],
    unwrappable.Unwrappable[_FirstType, None],
    equable.Equable,
):
    """
    Concrete interface for ``Maybe`` type.

    Can be unwrapped and compared.
    """

    __slots__ = ()

    @abstractmethod
    def or_else_call(
        self,
        function: Callable[[], _ValueType],
    ) -> _FirstType | _ValueType:
        """Calls a function in case there nothing to unwrap."""


#: Type alias for kinds with two type arguments.
MaybeBased2 = MaybeBasedN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
MaybeBased3 = MaybeBasedN[_FirstType, _SecondType, _ThirdType]



from copy import copy, deepcopy

import pytest

from returns.context import RequiresContextResult
from returns.primitives.exceptions import ImmutableStateError


def test_immutable_copy():
    """Ensures that helper returns it self when passed to copy function."""
    context_result = RequiresContextResult.from_value(1)
    assert context_result is copy(context_result)


def test_immutable_deepcopy():
    """Ensures that helper returns it self when passed to deepcopy function."""
    context_result = RequiresContextResult.from_value(1)
    assert context_result is deepcopy(context_result)


def test_requires_context_result_immutable():
    """Ensures that container is immutable."""
    with pytest.raises(ImmutableStateError):
        RequiresContextResult.from_value(1).abc = 1



from dataclasses import dataclass
from typing import Final

from returns.maybe import Maybe, Nothing, Some


@dataclass
class _Book:
    book_id: int
    name: str


_BOOK_LIST: Final = (
    _Book(book_id=1, name='Category Theory for Programmers'),
    _Book(book_id=2, name='Fluent Python'),
    _Book(book_id=3, name='Learn You Some Erlang for Great Good'),
    _Book(book_id=4, name='Learn You a Haskell for Great Good'),
)


def _find_book(book_id: int) -> Maybe[_Book]:
    for book in _BOOK_LIST:
        if book.book_id == book_id:
            return Some(book)
    return Nothing


if __name__ == '__main__':
    desired_book = _find_book(2)
    match desired_book:
        # Matches any `Some` instance that contains a book named `Fluent Python`
        case Some(_Book(name='Fluent Python')):
            print('"Fluent Python" was found')

        # Matches any `Some` instance and binds its value to the `book` variable
        case Some(book):
            print(f'Book found: {book.name}')

        # Matches `Nothing` instance
        case Maybe.empty:
            print('Not found the desired book!')



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, TypeVar

from returns.primitives.hkt import Kind3, dekind
from returns.result import Result, Success

if TYPE_CHECKING:
    from returns.context import RequiresContextFutureResult

_ValueType_co = TypeVar('_ValueType_co', covariant=True)
_NewValueType = TypeVar('_NewValueType')
_ErrorType_co = TypeVar('_ErrorType_co', covariant=True)
_EnvType = TypeVar('_EnvType')


async def async_bind_async(
    function: Callable[
        [_ValueType_co],
        Awaitable[
            Kind3[
                RequiresContextFutureResult,
                _NewValueType,
                _ErrorType_co,
                _EnvType,
            ],
        ],
    ],
    container: RequiresContextFutureResult[
        _ValueType_co, _ErrorType_co, _EnvType
    ],
    deps: _EnvType,
) -> Result[_NewValueType, _ErrorType_co]:
    """Async binds a coroutine with container over a value."""
    inner_value = await container(deps)._inner_value  # noqa: SLF001
    if isinstance(inner_value, Success):
        return await dekind(  # noqa: SLF001
            await function(inner_value.unwrap()),
        )(deps)._inner_value
    return inner_value  # type: ignore[return-value]


async def async_compose_result(
    function: Callable[
        [Result[_ValueType_co, _ErrorType_co]],
        Kind3[
            RequiresContextFutureResult,
            _NewValueType,
            _ErrorType_co,
            _EnvType,
        ],
    ],
    container: RequiresContextFutureResult[
        _ValueType_co, _ErrorType_co, _EnvType
    ],
    deps: _EnvType,
) -> Result[_NewValueType, _ErrorType_co]:
    """Async composes ``Result`` based function."""
    new_container = dekind(function((await container(deps))._inner_value))  # noqa: SLF001
    return (await new_container(deps))._inner_value  # noqa: SLF001



from abc import ABC
from collections.abc import Callable, Generator, Iterator
from functools import wraps
from typing import TYPE_CHECKING, Any, ClassVar, Never, Optional, TypeVar, final

from typing_extensions import ParamSpec

from returns.interfaces.specific.maybe import MaybeBased2
from returns.primitives.container import BaseContainer, container_equality
from returns.primitives.exceptions import UnwrapFailedError
from returns.primitives.hkt import Kind1, SupportsKind1

# Definitions:
_ValueType_co = TypeVar('_ValueType_co', covariant=True)
_NewValueType = TypeVar('_NewValueType')

_FuncParams = ParamSpec('_FuncParams')


class Maybe(  # type: ignore[type-var]
    BaseContainer,
    SupportsKind1['Maybe', _ValueType_co],
    MaybeBased2[_ValueType_co, None],
    ABC,
):
    """
    Represents a result of a series of computations that can return ``None``.

    An alternative to using exceptions or constant ``is None`` checks.
    ``Maybe`` is an abstract type and should not be instantiated directly.
    Instead use ``Some`` and ``Nothing``.

    See also:
        - https://github.com/gcanti/fp-ts/blob/master/docs/modules/Option.ts.md

    """

    __slots__ = ()

    _inner_value: _ValueType_co | None
    __match_args__ = ('_inner_value',)

    #: Alias for `Nothing`
    empty: ClassVar['Maybe[Any]']

    #: Typesafe equality comparison with other `Result` objects.
    equals = container_equality

    def map(
        self,
        function: Callable[[_ValueType_co], _NewValueType],
    ) -> 'Maybe[_NewValueType]':
        """
        Composes successful container with a pure function.

        .. code:: python

          >>> from returns.maybe import Some, Nothing
          >>> def mappable(string: str) -> str:
          ...      return string + 'b'

          >>> assert Some('a').map(mappable) == Some('ab')
          >>> assert Nothing.map(mappable) == Nothing

        """

    def apply(
        self,
        function: Kind1['Maybe', Callable[[_ValueType_co], _NewValueType]],
    ) -> 'Maybe[_NewValueType]':
        """
        Calls a wrapped function in a container on this container.

        .. code:: python

          >>> from returns.maybe import Some, Nothing

          >>> def appliable(string: str) -> str:
          ...      return string + 'b'

          >>> assert Some('a').apply(Some(appliable)) == Some('ab')
          >>> assert Some('a').apply(Nothing) == Nothing
          >>> assert Nothing.apply(Some(appliable)) == Nothing
          >>> assert Nothing.apply(Nothing) == Nothing

        """

    def bind(
        self,
        function: Callable[[_ValueType_co], Kind1['Maybe', _NewValueType]],
    ) -> 'Maybe[_NewValueType]':
        """
        Composes successful container with a function that returns a container.

        .. code:: python

          >>> from returns.maybe import Nothing, Maybe, Some
          >>> def bindable(string: str) -> Maybe[str]:
          ...      return Some(string + 'b')

          >>> assert Some('a').bind(bindable) == Some('ab')
          >>> assert Nothing.bind(bindable) == Nothing

        """

    def bind_optional(
        self,
        function: Callable[[_ValueType_co], _NewValueType | None],
    ) -> 'Maybe[_NewValueType]':
        """
        Binds a function returning an optional value over a container.

        .. code:: python

          >>> from returns.maybe import Some, Nothing
          >>> from typing import Optional

          >>> def bindable(arg: str) -> Optional[int]:
          ...     return len(arg) if arg else None

          >>> assert Some('a').bind_optional(bindable) == Some(1)
          >>> assert Some('').bind_optional(bindable) == Nothing

        """

    def lash(
        self,
        function: Callable[[Any], Kind1['Maybe', _ValueType_co]],
    ) -> 'Maybe[_ValueType_co]':
        """
        Composes failed container with a function that returns a container.

        .. code:: python

          >>> from returns.maybe import Maybe, Some, Nothing

          >>> def lashable(arg=None) -> Maybe[str]:
          ...      return Some('b')

          >>> assert Some('a').lash(lashable) == Some('a')
          >>> assert Nothing.lash(lashable) == Some('b')

        We need this feature to make ``Maybe`` compatible
        with different ``Result`` like operations.

        """

    def __iter__(self) -> Iterator[_ValueType_co]:
        """API for :ref:`do-notation`."""
        yield self.unwrap()

    @classmethod
    def do(
        cls,
        expr: Generator[_NewValueType, None, None],
    ) -> 'Maybe[_NewValueType]':
        """
        Allows working with unwrapped values of containers in a safe way.

        .. code:: python

          >>> from returns.maybe import Maybe, Some, Nothing

          >>> assert Maybe.do(
          ...     first + second
          ...     for first in Some(2)
          ...     for second in Some(3)
          ... ) == Some(5)

          >>> assert Maybe.do(
          ...     first + second
          ...     for first in Some(2)
          ...     for second in Nothing
          ... ) == Nothing

        See :ref:`do-notation` to learn more.

        """
        try:
            return Maybe.from_value(next(expr))
        except UnwrapFailedError as exc:
            return exc.halted_container  # type: ignore

    def value_or(
        self,
        default_value: _NewValueType,
    ) -> _ValueType_co | _NewValueType:
        """
        Get value from successful container or default value from failed one.

        .. code:: python

          >>> from returns.maybe import Nothing, Some
          >>> assert Some(0).value_or(1) == 0
          >>> assert Nothing.value_or(1) == 1

        """

    def or_else_call(
        self,
        function: Callable[[], _NewValueType],
    ) -> _ValueType_co | _NewValueType:
        """
        Get value from successful container or default value from failed one.

        Really close to :meth:`~Maybe.value_or` but works with lazy values.
        This method is unique to ``Maybe`` container, because other containers
        do have ``.alt`` method.

        But, ``Maybe`` does not have this method.
        There's nothing to ``alt`` in ``Nothing``.

        Instead, it has this method to execute
        some function if called on a failed container:

        .. code:: pycon

          >>> from returns.maybe import Some, Nothing
          >>> assert Some(1).or_else_call(lambda: 2) == 1
          >>> assert Nothing.or_else_call(lambda: 2) == 2

        It might be useful to work with exceptions as well:

        .. code:: pycon

          >>> def fallback() -> Never:
          ...    raise ValueError('Nothing!')

          >>> Nothing.or_else_call(fallback)
          Traceback (most recent call last):
            ...
          ValueError: Nothing!

        """

    def unwrap(self) -> _ValueType_co:
        """
        Get value from successful container or raise exception for failed one.

        .. code:: pycon
          :force:

          >>> from returns.maybe import Nothing, Some
          >>> assert Some(1).unwrap() == 1

          >>> Nothing.unwrap()
          Traceback (most recent call last):
            ...
          returns.primitives.exceptions.UnwrapFailedError

        """

    def failure(self) -> None:
        """
        Get failed value from failed container or raise exception from success.

        .. code:: pycon
          :force:

          >>> from returns.maybe import Nothing, Some
          >>> assert Nothing.failure() is None

          >>> Some(1).failure()
          Traceback (most recent call last):
            ...
          returns.primitives.exceptions.UnwrapFailedError

        """

    @classmethod
    def from_value(
        cls,
        inner_value: _NewValueType,
    ) -> 'Maybe[_NewValueType]':
        """
        Creates new instance of ``Maybe`` container based on a value.

        .. code:: python

          >>> from returns.maybe import Maybe, Some
          >>> assert Maybe.from_value(1) == Some(1)
          >>> assert Maybe.from_value(None) == Some(None)

        """
        return Some(inner_value)

    @classmethod
    def from_optional(
        cls,
        inner_value: _NewValueType | None,
    ) -> 'Maybe[_NewValueType]':
        """
        Creates new instance of ``Maybe`` container based on an optional value.

        .. code:: python

          >>> from returns.maybe import Maybe, Some, Nothing
          >>> assert Maybe.from_optional(1) == Some(1)
          >>> assert Maybe.from_optional(None) == Nothing

        """
        if inner_value is None:
            return _Nothing(inner_value)
        return Some(inner_value)

    def __bool__(self) -> bool:
        """Convert (or treat) an instance of ``Maybe`` as a boolean."""


@final
class _Nothing(Maybe[Any]):
    """Represents an empty state."""

    __slots__ = ()

    _inner_value: None
    _instance: Optional['_Nothing'] = None

    def __new__(cls, *args: Any, **kwargs: Any) -> '_Nothing':
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self, inner_value: None = None) -> None:  # noqa: WPS632
        """
        Private constructor for ``_Nothing`` type.

        Use :attr:`~Nothing` instead.
        Wraps the given value in the ``_Nothing`` container.

        ``inner_value`` can only be ``None``.
        """
        super().__init__(None)

    def __repr__(self):
        """
        Custom ``str`` definition without the state inside.

        .. code:: python

          >>> from returns.maybe import Nothing
          >>> assert str(Nothing) == '<Nothing>'
          >>> assert repr(Nothing) == '<Nothing>'

        """
        return '<Nothing>'

    def map(self, function):
        """Does nothing for ``Nothing``."""
        return self

    def apply(self, container):
        """Does nothing for ``Nothing``."""
        return self

    def bind(self, function):
        """Does nothing for ``Nothing``."""
        return self

    def bind_optional(self, function):
        """Does nothing."""
        return self

    def lash(self, function):
        """Composes this container with a function returning container."""
        return function(None)

    def value_or(self, default_value):
        """Returns default value."""
        return default_value

    def or_else_call(self, function):
        """Returns the result of a passed function."""
        return function()

    def unwrap(self):
        """Raises an exception, since it does not have a value inside."""
        raise UnwrapFailedError(self)

    def failure(self) -> None:
        """Returns failed value."""
        return self._inner_value

    def __bool__(self):
        """Returns ``False``."""
        return False


@final
class Some(Maybe[_ValueType_co]):
    """
    Represents a calculation which has succeeded and contains the value.

    Quite similar to ``Success`` type.
    """

    __slots__ = ()

    _inner_value: _ValueType_co

    def __init__(self, inner_value: _ValueType_co) -> None:
        """Some constructor."""
        super().__init__(inner_value)

    if not TYPE_CHECKING:  # noqa: WPS604  # pragma: no branch

        def bind(self, function):
            """Binds current container to a function that returns container."""
            return function(self._inner_value)

        def bind_optional(self, function):
            """Binds a function returning an optional value over a container."""
            return Maybe.from_optional(function(self._inner_value))

        def unwrap(self):
            """Returns inner value for successful container."""
            return self._inner_value

    def map(self, function):
        """Composes current container with a pure function."""
        return Some(function(self._inner_value))

    def apply(self, container):
        """Calls a wrapped function in a container on this container."""
        if isinstance(container, Some):
            return self.map(container.unwrap())  # type: ignore
        return container

    def lash(self, function):
        """Does nothing for ``Some``."""
        return self

    def value_or(self, default_value):
        """Returns inner value for successful container."""
        return self._inner_value

    def or_else_call(self, function):
        """Returns inner value for successful container."""
        return self._inner_value

    def failure(self):
        """Raises exception for successful container."""
        raise UnwrapFailedError(self)

    def __bool__(self):
        """
        Returns ``True```.

        Any instance of ``Something`` is treated
        as ``True``, even ``Something(None)``.
        """
        return True


#: Public unit value of protected :class:`~_Nothing` type.
Nothing: Maybe[Never] = _Nothing()
Maybe.empty = Nothing


def maybe(
    function: Callable[_FuncParams, _ValueType_co | None],
) -> Callable[_FuncParams, Maybe[_ValueType_co]]:
    """
    Decorator to convert ``None``-returning function to ``Maybe`` container.

    This decorator works with sync functions only. Example:

    .. code:: python

      >>> from typing import Optional
      >>> from returns.maybe import Nothing, Some, maybe

      >>> @maybe
      ... def might_be_none(arg: int) -> Optional[int]:
      ...     if arg == 0:
      ...         return None
      ...     return 1 / arg

      >>> assert might_be_none(0) == Nothing
      >>> assert might_be_none(1) == Some(1.0)

    """

    @wraps(function)
    def decorator(
        *args: _FuncParams.args,
        **kwargs: _FuncParams.kwargs,
    ) -> Maybe[_ValueType_co]:
        return Maybe.from_optional(function(*args, **kwargs))

    return decorator



from returns.context import (
    ReaderResult,
    ReaderResultE,
    RequiresContextResult,
    RequiresContextResultE,
)


def _function(arg: int) -> RequiresContextResultE[float, int]:
    if arg == 0:
        return RequiresContextResult.from_failure(
            ZeroDivisionError('Divided by 0'),
        )
    return RequiresContextResult.from_value(10 / arg)


def test_requires_context_resulte():
    """Ensures that RequiresContextResultE correctly typecast."""
    container: RequiresContextResult[float, Exception, int] = _function(1)
    assert container(0) == RequiresContextResult.from_value(10.0)(0)


def test_requires_context_aliases():
    """Ensures that ReaderResult correctly typecast."""
    container: ReaderResultE[float, int] = _function(1)
    container2: ReaderResult[float, Exception, int] = _function(1)
    container3: ReaderResultE[float, int] = ReaderResultE.from_value(
        10.0,
    )
    container4: ReaderResultE[float, int] = ReaderResult.from_value(10.0)

    assert container(0) == container2(0) == container3(0) == container4(0)
    assert container(0) == RequiresContextResult.from_value(10.0)(0)



from collections.abc import Callable, Sequence
from typing import ClassVar, Never, TypeVar, final

from returns.interfaces import applicative, bindable
from returns.primitives.asserts import assert_equal
from returns.primitives.hkt import KindN
from returns.primitives.laws import (
    Law,
    Law1,
    Law3,
    Lawful,
    LawSpecDef,
    law_definition,
)

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')

# Only used in laws:
_NewType1 = TypeVar('_NewType1')
_NewType2 = TypeVar('_NewType2')


@final
class _LawSpec(LawSpecDef):
    """
    Container laws.

    Definition: https://wiki.haskell.org/Monad_laws
    Good explanation: https://bit.ly/2Qsi5re
    """

    __slots__ = ()

    @law_definition
    def left_identity_law(
        raw_value: _FirstType,
        container: 'ContainerN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[
            [_FirstType],
            KindN['ContainerN', _NewType1, _SecondType, _ThirdType],
        ],
    ) -> None:
        """
        Left identity.

        The first law states that if we take a value, put it in a default
        context with return and then feed it to a function by using ``bind``,
        it's the same as just taking the value and applying the function to it.
        """
        assert_equal(
            container.from_value(raw_value).bind(function),
            function(raw_value),
        )

    @law_definition
    def right_identity_law(
        container: 'ContainerN[_FirstType, _SecondType, _ThirdType]',
    ) -> None:
        """
        Right identity.

        The second law states that if we have a container value
        and we use ``bind`` to feed it to ``.from_value``,
        the result is our original container value.
        """
        assert_equal(
            container,
            container.bind(
                container.from_value,
            ),
        )

    @law_definition
    def associative_law(
        container: 'ContainerN[_FirstType, _SecondType, _ThirdType]',
        first: Callable[
            [_FirstType],
            KindN['ContainerN', _NewType1, _SecondType, _ThirdType],
        ],
        second: Callable[
            [_NewType1],
            KindN['ContainerN', _NewType2, _SecondType, _ThirdType],
        ],
    ) -> None:
        """
        Associativity law.

        The final monad law says that when
        we have a chain of container functions applications with ``bind``,
        it shouldn't matter how they're nested.
        """
        assert_equal(
            container.bind(first).bind(second),
            container.bind(lambda inner: first(inner).bind(second)),
        )


class ContainerN(
    applicative.ApplicativeN[_FirstType, _SecondType, _ThirdType],
    bindable.BindableN[_FirstType, _SecondType, _ThirdType],
    Lawful['ContainerN[_FirstType, _SecondType, _ThirdType]'],
):
    """
    Handy alias for types with ``.bind``, ``.map``, and ``.apply`` methods.

    Should be a base class for almost any containers you write.

    See also:
        - https://bit.ly/2CTEVov

    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (
        Law3(_LawSpec.left_identity_law),
        Law1(_LawSpec.right_identity_law),
        Law3(_LawSpec.associative_law),
    )


#: Type alias for kinds with one type argument.
Container1 = ContainerN[_FirstType, Never, Never]

#: Type alias for kinds with two type arguments.
Container2 = ContainerN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
Container3 = ContainerN[_FirstType, _SecondType, _ThirdType]



"""
Used to register all our types as hypothesis strategies.

See: https://hypothesis.readthedocs.io/en/latest/strategies.html

But, beware that we only register concrete types here,
interfaces won't be registered!

"""

from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from returns.primitives.laws import Lawful

_Inst = TypeVar('_Inst', bound='Lawful')


def _setup_hook() -> None:
    from hypothesis import strategies as st  # noqa: PLC0415

    from returns.context import (  # noqa: PLC0415
        RequiresContext,
        RequiresContextFutureResult,
        RequiresContextIOResult,
        RequiresContextResult,
    )
    from returns.future import Future, FutureResult  # noqa: PLC0415
    from returns.io import IO, IOResult  # noqa: PLC0415
    from returns.maybe import Maybe  # noqa: PLC0415
    from returns.result import Result  # noqa: PLC0415

    def factory(
        container_type: type[_Inst],
    ) -> Callable[[Any], st.SearchStrategy[_Inst]]:
        def decorator(thing: Any) -> st.SearchStrategy[_Inst]:
            from returns.contrib.hypothesis.containers import (  # noqa: PLC0415
                strategy_from_container,
            )

            return strategy_from_container(container_type)(thing)

        return decorator

    #: Our types that we register in hypothesis
    #: to be working with ``st.from_type``
    registered_types: Sequence[type[Lawful]] = (
        Result,
        Maybe,
        IO,
        IOResult,
        Future,
        FutureResult,
        RequiresContext,
        RequiresContextResult,
        RequiresContextIOResult,
        RequiresContextFutureResult,
    )

    for type_ in registered_types:
        st.register_type_strategy(type_, factory(type_))



from abc import abstractmethod
from collections.abc import Callable, Iterable
from typing import TypeVar, final

from returns.interfaces.applicative import ApplicativeN
from returns.interfaces.failable import FailableN
from returns.primitives.hkt import KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_ApplicativeKind = TypeVar('_ApplicativeKind', bound=ApplicativeN)
_FailableKind = TypeVar('_FailableKind', bound=FailableN)


class AbstractFold:
    """
    A collection of different helpers to write declarative ``Iterable`` actions.

    Allows to work with iterables.

    .. rubric:: Implementation

    ``AbstractFold`` and ``Fold`` types are special.
    They have double definition for each method: public and protected ones.
    Why?

    Because you cannot override ``@kinded`` method due to a ``mypy`` bug.
    So, there are two opportunities for us here:

    1. Declare all method as ``@final`` and do not allow to change anything
    2. Use delegation to protected unkinded methods

    We have chosen the second way! Here's how it works:

    1. Public methods are ``@kinded`` for better typing and cannot be overridden
    2. Protected methods are unkinded and can be overridden in subtyping

    Now, if you need to make a change into our implementation,
    then you can subclass ``Fold`` or ``AbstractFold`` and then
    change an implementation of any unkinded protected method.
    """

    __slots__ = ()

    @final
    @kinded
    @classmethod
    def loop(
        cls,
        iterable: Iterable[
            KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
        ],
        acc: KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType],
        function: Callable[
            [_FirstType],
            Callable[[_UpdatedType], _UpdatedType],
        ],
    ) -> KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType]:
        """
        Allows to make declarative loops for any ``ApplicativeN`` subtypes.

        Quick example:

        .. code:: python

          >>> from typing import Callable
          >>> from returns.maybe import Some
          >>> from returns.iterables import Fold

          >>> def sum_two(first: int) -> Callable[[int], int]:
          ...     return lambda second: first + second

          >>> assert Fold.loop(
          ...     [Some(1), Some(2), Some(3)],
          ...     Some(10),
          ...     sum_two,
          ... ) == Some(16)

        Looks like ``foldl`` in some other languages with some more specifics.
        See: https://philipschwarz.dev/fpilluminated/?page_id=348#bwg3/137

        .. image:: https://i.imgur.com/Tza1isS.jpg

        Is also quite similar to ``reduce``.

        Public interface for ``_loop`` method. Cannot be modified directly.
        """
        return cls._loop(iterable, acc, function, _concat_applicative)

    @final
    @kinded
    @classmethod
    def collect(
        cls,
        iterable: Iterable[
            KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
        ],
        acc: KindN[
            _ApplicativeKind,
            'tuple[_FirstType, ...]',
            _SecondType,
            _ThirdType,
        ],
    ) -> KindN[
        _ApplicativeKind,
        'tuple[_FirstType, ...]',
        _SecondType,
        _ThirdType,
    ]:
        """
        Transforms an iterable of containers into a single container.

        Quick example for regular containers:

        .. code:: python

          >>> from returns.io import IO
          >>> from returns.iterables import Fold

          >>> items = [IO(1), IO(2)]
          >>> assert Fold.collect(items, IO(())) == IO((1, 2))

        If container can have failed values,
        then this strategy fails on any existing failed like type.

        It is enough to have even a single failed value in iterable
        for this type to convert the whole operation result to be a failure.
        Let's see how it works:

        .. code:: python

          >>> from returns.result import Success, Failure
          >>> from returns.iterables import Fold

          >>> empty = []
          >>> all_success = [Success(1), Success(2), Success(3)]
          >>> has_failure = [Success(1), Failure('a'), Success(3)]
          >>> all_failures = [Failure('a'), Failure('b')]

          >>> acc = Success(())  # empty tuple

          >>> assert Fold.collect(empty, acc) == Success(())
          >>> assert Fold.collect(all_success, acc) == Success((1, 2, 3))
          >>> assert Fold.collect(has_failure, acc) == Failure('a')
          >>> assert Fold.collect(all_failures, acc) == Failure('a')

        If that's now what you need, check out
        :meth:`~AbstractFold.collect_all`
        to force collect all non-failed values.

        Public interface for ``_collect`` method. Cannot be modified directly.
        """
        return cls._collect(iterable, acc)

    @final
    @kinded
    @classmethod
    def collect_all(
        cls,
        iterable: Iterable[
            KindN[_FailableKind, _FirstType, _SecondType, _ThirdType],
        ],
        acc: KindN[
            _FailableKind,
            'tuple[_FirstType, ...]',
            _SecondType,
            _ThirdType,
        ],
    ) -> KindN[
        _FailableKind,
        'tuple[_FirstType, ...]',
        _SecondType,
        _ThirdType,
    ]:
        """
        Transforms an iterable of containers into a single container.

        This method only works with ``FailableN`` subtypes,
        not just any ``ApplicativeN`` like :meth:`~AbstractFold.collect`.

        Strategy to extract all successful values
        even if there are failed values.

        If there's at least one successful value
        and any amount of failed values,
        we will still return all collected successful values.

        We can return failed value for this strategy only in a single case:
        when default element is a failed value.

        Let's see how it works:

        .. code:: python

          >>> from returns.result import Success, Failure
          >>> from returns.iterables import Fold

          >>> empty = []
          >>> all_success = [Success(1), Success(2), Success(3)]
          >>> has_failure = [Success(1), Failure('a'), Success(3)]
          >>> all_failures = [Failure('a'), Failure('b')]

          >>> acc = Success(())  # empty tuple

          >>> assert Fold.collect_all(empty, acc) == Success(())
          >>> assert Fold.collect_all(all_success, acc) == Success((1, 2, 3))
          >>> assert Fold.collect_all(has_failure, acc) == Success((1, 3))
          >>> assert Fold.collect_all(all_failures, acc) == Success(())
          >>> assert Fold.collect_all(empty, Failure('c')) == Failure('c')

        If that's now what you need, check out :meth:`~AbstractFold.collect`
        to collect only successful values and fail on any failed ones.

        Public interface for ``_collect_all`` method.
        Cannot be modified directly.
        """
        return cls._collect_all(iterable, acc)

    # Protected part
    # ==============

    @classmethod
    @abstractmethod
    def _loop(
        cls,
        iterable: Iterable[
            KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
        ],
        acc: KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType],
        function: Callable[
            [_FirstType],
            Callable[[_UpdatedType], _UpdatedType],
        ],
        concat: Callable[
            [
                KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
                KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType],
                KindN[
                    _ApplicativeKind,
                    Callable[
                        [_FirstType],
                        Callable[[_UpdatedType], _UpdatedType],
                    ],
                    _SecondType,
                    _ThirdType,
                ],
            ],
            KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType],
        ],
    ) -> KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType]:
        """
        Protected part of ``loop`` method.

        Can be replaced in subclasses for better performance, etc.
        """

    @classmethod
    def _collect(
        cls,
        iterable: Iterable[
            KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
        ],
        acc: KindN[
            _ApplicativeKind,
            'tuple[_FirstType, ...]',
            _SecondType,
            _ThirdType,
        ],
    ) -> KindN[
        _ApplicativeKind,
        'tuple[_FirstType, ...]',
        _SecondType,
        _ThirdType,
    ]:
        return cls._loop(
            iterable,
            acc,
            _concat_sequence,
            _concat_applicative,
        )

    @classmethod
    def _collect_all(
        cls,
        iterable: Iterable[
            KindN[_FailableKind, _FirstType, _SecondType, _ThirdType],
        ],
        acc: KindN[
            _FailableKind,
            'tuple[_FirstType, ...]',
            _SecondType,
            _ThirdType,
        ],
    ) -> KindN[
        _FailableKind,
        'tuple[_FirstType, ...]',
        _SecondType,
        _ThirdType,
    ]:
        return cls._loop(
            iterable,
            acc,
            _concat_sequence,
            _concat_failable_safely,
        )


class Fold(AbstractFold):
    """
    Concrete implementation of ``AbstractFold`` of end users.

    Use it by default.
    """

    __slots__ = ()

    @classmethod
    def _loop(
        cls,
        iterable: Iterable[
            KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
        ],
        acc: KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType],
        function: Callable[
            [_FirstType],
            Callable[[_UpdatedType], _UpdatedType],
        ],
        concat: Callable[
            [
                KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
                KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType],
                KindN[
                    _ApplicativeKind,
                    Callable[
                        [_FirstType],
                        Callable[[_UpdatedType], _UpdatedType],
                    ],
                    _SecondType,
                    _ThirdType,
                ],
            ],
            KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType],
        ],
    ) -> KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType]:
        """
        Protected part of ``loop`` method.

        Can be replaced in subclasses for better performance, etc.
        """
        wrapped = acc.from_value(function)
        for current in iterable:
            acc = concat(current, acc, wrapped)
        return acc


# Helper functions
# ================


def _concat_sequence(
    first: _FirstType,
) -> Callable[
    ['tuple[_FirstType, ...]'],
    'tuple[_FirstType, ...]',
]:
    """
    Concats a given item to an existing sequence.

    We use explicit curring with ``lambda`` function because,
    ``@curry`` decorator is way slower. And we don't need its features here.
    But, your functions can use ``@curry`` if you need it.
    """
    return lambda second: (*second, first)


def _concat_applicative(
    current: KindN[
        _ApplicativeKind,
        _FirstType,
        _SecondType,
        _ThirdType,
    ],
    acc: KindN[
        _ApplicativeKind,
        _UpdatedType,
        _SecondType,
        _ThirdType,
    ],
    function: KindN[
        _ApplicativeKind,
        Callable[[_FirstType], Callable[[_UpdatedType], _UpdatedType]],
        _SecondType,
        _ThirdType,
    ],
) -> KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType]:
    """Concats two applicatives using a curried-like function."""
    return acc.apply(current.apply(function))


def _concat_failable_safely(
    current: KindN[
        _FailableKind,
        _FirstType,
        _SecondType,
        _ThirdType,
    ],
    acc: KindN[
        _FailableKind,
        _UpdatedType,
        _SecondType,
        _ThirdType,
    ],
    function: KindN[
        _FailableKind,
        Callable[[_FirstType], Callable[[_UpdatedType], _UpdatedType]],
        _SecondType,
        _ThirdType,
    ],
) -> KindN[_FailableKind, _UpdatedType, _SecondType, _ThirdType]:
    """
    Concats two ``FailableN`` using a curried-like function and a fallback.

    We need both ``.apply`` and ``.lash`` methods here.
    """
    return _concat_applicative(current, acc, function).lash(lambda _: acc)



from collections.abc import Callable
from typing import TypeVar

from returns.contrib.hypothesis.laws import check_all_laws
from returns.interfaces import applicative
from returns.primitives.container import BaseContainer
from returns.primitives.hkt import Kind1, SupportsKind1

_ValueType = TypeVar('_ValueType')
_NewValueType = TypeVar('_NewValueType')


class _Wrapper(
    BaseContainer,
    SupportsKind1['_Wrapper', _ValueType],
    applicative.Applicative1[_ValueType],
):
    _inner_value: _ValueType

    def __init__(self, inner_value: _ValueType) -> None:
        super().__init__(inner_value)

    def map(
        self,
        function: Callable[[_ValueType], _NewValueType],
    ) -> '_Wrapper[_NewValueType]':
        return _Wrapper(function(self._inner_value))

    def apply(
        self,
        container: Kind1['_Wrapper', Callable[[_ValueType], _NewValueType]],
    ) -> '_Wrapper[_NewValueType]':
        function = container._inner_value  # noqa: SLF001
        return _Wrapper(function(self._inner_value))

    @classmethod
    def from_value(
        cls,
        inner_value: _NewValueType,
    ) -> '_Wrapper[_NewValueType]':
        return _Wrapper(inner_value)


check_all_laws(_Wrapper)



"""
This module is special.

``Reader`` does not produce ``ReaderLikeN`` interface as other containers.

Because ``Reader`` can be used with two or three type arguments:
- ``RequiresContext[value, env]``
- ``RequiresContextResult[value, error, env]``

Because the second type argument changes its meaning
based on the used ``KindN`` instance,
we need to have two separate interfaces for two separate use-cases:
- ``ReaderLike2`` is used for types where the second type argument is ``env``
- ``ReaderLike3`` is used for types where the third type argument is ``env``

We also have two methods and two poinfree helpers
for ``bind_context`` composition: one for each interface.

Furthermore, ``Reader`` cannot have ``ReaderLike1`` type,
because we need both ``value`` and ``env`` types at all cases.

See also:
    - https://github.com/dry-python/returns/issues/485

"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar, final

from returns.interfaces.container import Container2, Container3
from returns.primitives.hkt import Kind2, Kind3
from returns.primitives.laws import (
    Law,
    Law2,
    Lawful,
    LawSpecDef,
    law_definition,
)

if TYPE_CHECKING:
    from returns.context import NoDeps, RequiresContext  # noqa: WPS433

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_ValueType = TypeVar('_ValueType')
_EnvType = TypeVar('_EnvType')

_ReaderLike2Type = TypeVar('_ReaderLike2Type', bound='ReaderLike2')
_ReaderLike3Type = TypeVar('_ReaderLike3Type', bound='ReaderLike3')


class Contextable(Generic[_ValueType, _EnvType]):
    """
    Special type we use as a base one for all callble ``Reader`` instances.

    It only has a single method.
    And is a base type for every single one of them.

    But, each ``Reader`` defines the return type differently.
    For example:

    - ``Reader`` has just ``_ReturnType``
    - ``ReaderResult`` has ``Result[_FirstType, _SecondType]``
    - ``ReaderIOResult`` has ``IOResult[_FirstType, _SecondType]``

    And so on.
    """

    __slots__ = ()

    @abstractmethod
    def __call__(self, deps: _EnvType) -> _ValueType:
        """Receives one parameter, returns a value. As simple as that."""


class ReaderLike2(Container2[_FirstType, _SecondType]):
    """
    Reader interface for ``Kind2`` based types.

    It has two type arguments and treats the second type argument as env type.
    """

    __slots__ = ()

    @property
    @abstractmethod
    def no_args(self: _ReaderLike2Type) -> NoDeps:
        """Is required to call ``Reader`` with no explicit arguments."""

    @abstractmethod
    def bind_context(
        self: _ReaderLike2Type,
        function: Callable[
            [_FirstType],
            RequiresContext[_UpdatedType, _SecondType],
        ],
    ) -> Kind2[_ReaderLike2Type, _UpdatedType, _SecondType]:
        """Allows to apply a wrapped function over a ``Reader`` container."""

    @abstractmethod
    def modify_env(
        self: _ReaderLike2Type,
        function: Callable[[_UpdatedType], _SecondType],
    ) -> Kind2[_ReaderLike2Type, _FirstType, _UpdatedType]:
        """Transforms the environment before calling the container."""

    @classmethod
    @abstractmethod
    def ask(
        cls: type[_ReaderLike2Type],
    ) -> Kind2[_ReaderLike2Type, _SecondType, _SecondType]:
        """Returns the dependencies inside the container."""

    @classmethod
    @abstractmethod
    def from_context(
        cls: type[_ReaderLike2Type],
        inner_value: RequiresContext[_ValueType, _EnvType],
    ) -> Kind2[_ReaderLike2Type, _ValueType, _EnvType]:
        """Unit method to create new containers from successful ``Reader``."""


class CallableReader2(
    ReaderLike2[_FirstType, _SecondType],
    Contextable[_ValueType, _EnvType],
):
    """
    Intermediate interface for ``ReaderLike2`` + ``__call__`` method.

    Has 4 type variables to type ``Reader`` and ``__call__`` independently.
    Since, we don't have any other fancy ways of doing it.

    Should not be used directly
    other than defining your own ``Reader`` interfaces.
    """

    __slots__ = ()


class ReaderLike3(Container3[_FirstType, _SecondType, _ThirdType]):
    """
    Reader interface for ``Kind3`` based types.

    It has three type arguments and treats the third type argument as env type.
    The second type argument is not used here.
    """

    __slots__ = ()

    @property
    @abstractmethod
    def no_args(self: _ReaderLike3Type) -> NoDeps:
        """Is required to call ``Reader`` with no explicit arguments."""

    @abstractmethod
    def bind_context(
        self: _ReaderLike3Type,
        function: Callable[
            [_FirstType],
            RequiresContext[_UpdatedType, _ThirdType],
        ],
    ) -> Kind3[_ReaderLike3Type, _UpdatedType, _SecondType, _ThirdType]:
        """Allows to apply a wrapped function over a ``Reader`` container."""

    @abstractmethod
    def modify_env(
        self: _ReaderLike3Type,
        function: Callable[[_UpdatedType], _ThirdType],
    ) -> Kind3[_ReaderLike3Type, _FirstType, _SecondType, _UpdatedType]:
        """Transforms the environment before calling the container."""

    @classmethod
    @abstractmethod
    def ask(
        cls: type[_ReaderLike3Type],
    ) -> Kind3[_ReaderLike3Type, _ThirdType, _SecondType, _ThirdType]:
        """Returns the dependencies inside the container."""

    @classmethod
    @abstractmethod
    def from_context(
        cls: type[_ReaderLike3Type],
        inner_value: RequiresContext[_ValueType, _EnvType],
    ) -> Kind3[_ReaderLike3Type, _ValueType, _SecondType, _EnvType]:
        """Unit method to create new containers from successful ``Reader``."""


class CallableReader3(
    ReaderLike3[_FirstType, _SecondType, _ThirdType],
    Contextable[_ValueType, _EnvType],
):
    """
    Intermediate interface for ``ReaderLike3`` + ``__call__`` method.

    Has 5 type variables to type ``Reader`` and ``__call__`` independently.
    Since, we don't have any other fancy ways of doing it.

    Should not be used directly
    other than defining your own ``Reader`` interfaces.
    """

    __slots__ = ()


@final
class _LawSpec(LawSpecDef):
    """
    Concrete laws for ``ReaderBased2``.

    See: https://github.com/haskell/mtl/pull/61/files
    """

    __slots__ = ()

    @law_definition
    def purity_law(
        container: ReaderBased2[_FirstType, _SecondType],
        env: _SecondType,
    ) -> None:
        """Calling a ``Reader`` twice has the same result with the same env."""
        assert container(env) == container(env)

    @law_definition
    def asking_law(
        container: ReaderBased2[_FirstType, _SecondType],
        env: _SecondType,
    ) -> None:
        """Asking for an env, always returns the env."""
        assert container.ask().__call__(  # noqa: PLC2801
            env,
        ) == container.from_value(env).__call__(env)  # noqa: PLC2801


class ReaderBased2(
    CallableReader2[
        _FirstType,
        _SecondType,
        # Used for call typing:
        _FirstType,
        _SecondType,
    ],
    Lawful['ReaderBased2[_FirstType, _SecondType]'],
):
    """
    This interface is very specific to our ``Reader`` type.

    The only thing that differs from ``ReaderLike2`` is that we know
    the specific types for its ``__call__`` method.
    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (
        Law2(_LawSpec.purity_law),
        Law2(_LawSpec.asking_law),
    )



from returns.result import Failure, Result, Success


def test_bind():
    """Ensures that bind works."""

    def factory(inner_value: int) -> Result[int, str]:
        if inner_value > 0:
            return Success(inner_value * 2)
        return Failure(str(inner_value))

    input_value = 5
    bound: Result[int, str] = Success(input_value)

    assert bound.bind(factory) == factory(input_value)
    assert Success(input_value).bind(factory) == factory(input_value)
    assert str(bound.bind(factory)) == '<Success: 10>'

    input_value = 0
    bound2: Result[int, str] = Success(input_value)

    assert bound2.bind(factory) == factory(input_value)
    assert str(bound2.bind(factory)) == '<Failure: 0>'


def test_left_identity_success():
    """Ensures that left identity works for Success container."""

    def factory(inner_value: int) -> Result[int, str]:
        return Success(inner_value * 2)

    input_value = 5
    bound: Result[int, str] = Success(input_value)

    assert bound.bind(factory) == factory(input_value)


def test_left_identity_failure():
    """Ensures that left identity works for Failure container."""

    def factory(inner_value: int) -> Result[int, int]:
        return Failure(6)

    input_value = 5
    bound: Result[int, int] = Failure(input_value)

    assert bound.bind(factory) == Failure(input_value)
    assert Failure(input_value).bind(factory) == Failure(5)
    assert str(bound) == '<Failure: 5>'


def test_lash_success():
    """Ensures that lash works for Success container."""

    def factory(inner_value) -> Result[int, str]:
        return Success(inner_value * 2)

    bound = Success(5).lash(factory)

    assert bound == Success(5)
    assert Success(5).lash(factory) == Success(5)
    assert str(bound) == '<Success: 5>'


def test_lash_failure():
    """Ensures that lash works for Failure container."""

    def factory(inner_value: int) -> Result[str, int]:
        return Failure(inner_value + 1)

    expected = 6
    bound: Result[str, int] = Failure(5)

    assert bound.lash(factory) == Failure(expected)
    assert Failure(5).lash(factory) == Failure(expected)
    assert str(bound.lash(factory)) == '<Failure: 6>'



import pytest

from returns.functions import raise_exception
from returns.result import Failure, Success


class _CustomError(Exception):
    """Just for the test."""


@pytest.mark.parametrize(
    'exception_type',
    [
        TypeError,
        ValueError,
        _CustomError,
    ],
)
def test_raise_regular_exception(exception_type: type[Exception]):
    """Ensures that regular exception can be thrown."""
    with pytest.raises(exception_type):
        raise_exception(exception_type())


def test_failure_can_be_alted():
    """Ensures that exceptions can work with Failures."""
    failure = Failure(ValueError('Message'))
    with pytest.raises(ValueError, match='Message'):
        failure.alt(raise_exception)


def test_success_is_not_touched():
    """Ensures that exceptions can work with Success."""
    assert Success(1).alt(raise_exception) == Success(1)



"""
Custom mypy plugin to solve the temporary problem with python typing.

Important: we don't do anything ugly here.
We only solve problems of the current typing implementation.

``mypy`` API docs are here:
https://mypy.readthedocs.io/en/latest/extending_mypy.html

We use ``pytest-mypy-plugins`` to test that it works correctly, see:
https://github.com/mkurnikov/pytest-mypy-plugins
"""

from collections.abc import Callable, Mapping
from typing import ClassVar, TypeAlias, final

from mypy.plugin import (
    AttributeContext,
    FunctionContext,
    MethodContext,
    MethodSigContext,
    Plugin,
)
from mypy.types import CallableType
from mypy.types import Type as MypyType

from returns.contrib.mypy import _consts
from returns.contrib.mypy._features import (
    curry,
    do_notation,
    flow,
    kind,
    partial,
    pipe,
)

# Type aliases
# ============

#: Type for a function hook.
_FunctionCallback: TypeAlias = Callable[[FunctionContext], MypyType]

#: Type for attribute hook.
_AttributeCallback: TypeAlias = Callable[[AttributeContext], MypyType]

#: Type for a method hook.
_MethodCallback: TypeAlias = Callable[[MethodContext], MypyType]

#: Type for a method signature hook.
_MethodSigCallback: TypeAlias = Callable[[MethodSigContext], CallableType]


# Interface
# =========


@final
class _ReturnsPlugin(Plugin):
    """Our main plugin to dispatch different callbacks to specific features."""

    _function_hook_plugins: ClassVar[Mapping[str, _FunctionCallback]] = {
        _consts.TYPED_PARTIAL_FUNCTION: partial.analyze,
        _consts.TYPED_CURRY_FUNCTION: curry.analyze,
        _consts.TYPED_FLOW_FUNCTION: flow.analyze,
        _consts.TYPED_PIPE_FUNCTION: pipe.analyze,
        _consts.TYPED_KIND_DEKIND: kind.dekind,
    }

    _method_sig_hook_plugins: ClassVar[Mapping[str, _MethodSigCallback]] = {
        _consts.TYPED_PIPE_METHOD: pipe.signature,
        _consts.TYPED_KIND_KINDED_CALL: kind.kinded_signature,
    }

    _method_hook_plugins: ClassVar[Mapping[str, _MethodCallback]] = {
        _consts.TYPED_PIPE_METHOD: pipe.infer,
        _consts.TYPED_KIND_KINDED_CALL: kind.kinded_call,
        _consts.TYPED_KIND_KINDED_GET: kind.kinded_get_descriptor,
        **dict.fromkeys(_consts.DO_NOTATION_METHODS, do_notation.analyze),
    }

    def get_function_hook(
        self,
        fullname: str,
    ) -> _FunctionCallback | None:
        """
        Called for function return types from ``mypy``.

        Runs on each function call in the source code.
        We are only interested in a particular subset of all functions.
        So, we return a function handler for them.

        Otherwise, we return ``None``.
        """
        return self._function_hook_plugins.get(fullname)

    def get_attribute_hook(
        self,
        fullname: str,
    ) -> _AttributeCallback | None:
        """Called for any exiting or ``__getattr__`` attribute access."""
        if fullname.startswith(_consts.TYPED_KINDN_ACCESS):
            name_parts = fullname.split('.')
            attribute_name = name_parts[-1]
            if attribute_name.startswith('__') and attribute_name.endswith(
                '__'
            ):
                return None
            return kind.attribute_access
        return None

    def get_method_signature_hook(
        self,
        fullname: str,
    ) -> _MethodSigCallback | None:
        """Called for method signature from ``mypy``."""
        return self._method_sig_hook_plugins.get(fullname)

    def get_method_hook(
        self,
        fullname: str,
    ) -> _MethodCallback | None:
        """Called for method return types from ``mypy``."""
        return self._method_hook_plugins.get(fullname)


def plugin(version: str) -> type[Plugin]:
    """Plugin's public API and entrypoint."""
    return _ReturnsPlugin



import pytest
from test_hypothesis.test_laws import (
    test_custom_interface_with_laws,
    test_custom_type_applicative,
)

from returns.context import (
    RequiresContext,
    RequiresContextFutureResult,
    RequiresContextIOResult,
    RequiresContextResult,
)
from returns.future import Future, FutureResult
from returns.io import IO, IOResult
from returns.maybe import Maybe
from returns.primitives.laws import Law, Lawful
from returns.result import Result


@pytest.mark.parametrize(
    'container',
    [
        Result,
        Maybe,
        Future,
        FutureResult,
        IO,
        IOResult,
        RequiresContext,
        RequiresContextFutureResult,
        RequiresContextIOResult,
        RequiresContextResult,
    ],
)
def test_laws_resolution(container: type[Lawful]):
    """Ensures all tests are unique."""
    all_laws: list[Law] = []
    for laws in container.laws().values():
        all_laws.extend(laws)
    assert len(all_laws) == len(set(all_laws))


def test_container_defined_in_returns() -> None:
    """Check that it returns all interfaces for a container in `returns`."""
    result = Result.laws()

    assert sorted(str(interface) for interface in result) == [
        "<class 'returns.interfaces.altable.AltableN'>",
        "<class 'returns.interfaces.applicative.ApplicativeN'>",
        "<class 'returns.interfaces.container.ContainerN'>",
        "<class 'returns.interfaces.equable.Equable'>",
        "<class 'returns.interfaces.failable.DiverseFailableN'>",
        "<class 'returns.interfaces.failable.FailableN'>",
        "<class 'returns.interfaces.mappable.MappableN'>",
        "<class 'returns.interfaces.swappable.SwappableN'>",
    ]


def test_container_defined_outside_returns() -> None:
    """Check container defined outside `returns`."""
    result = test_custom_type_applicative._Wrapper.laws()  # noqa: SLF001

    assert sorted(str(interface) for interface in result) == [
        "<class 'returns.interfaces.applicative.ApplicativeN'>",
        "<class 'returns.interfaces.mappable.MappableN'>",
    ]


def test_interface_defined_outside_returns() -> None:
    """Check container with interface defined outside `returns`."""
    result = test_custom_interface_with_laws._Wrapper.laws()  # noqa: SLF001

    assert sorted(str(interface) for interface in result) == [
        (
            "<class 'test_hypothesis.test_laws.test_custom_interface_with_laws"
            "._MappableN'>"
        ),
    ]



from abc import abstractmethod
from collections.abc import Sequence
from typing import ClassVar, Never, TypeVar, final

from returns.interfaces import bimappable
from returns.primitives.asserts import assert_equal
from returns.primitives.hkt import KindN
from returns.primitives.laws import (
    Law,
    Law1,
    Lawful,
    LawSpecDef,
    law_definition,
)

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')

_SwappableType = TypeVar('_SwappableType', bound='SwappableN')


@final
class _LawSpec(LawSpecDef):
    """Laws for :class:`~SwappableN` type."""

    __slots__ = ()

    @law_definition
    def double_swap_law(
        container: 'SwappableN[_FirstType, _SecondType, _ThirdType]',
    ) -> None:
        """
        Swapping container twice.

        It ensure that we get the initial value back.
        In other words, swapping twice does nothing.
        """
        assert_equal(
            container,
            container.swap().swap(),
        )


class SwappableN(
    bimappable.BiMappableN[_FirstType, _SecondType, _ThirdType],
    Lawful['SwappableN[_FirstType, _SecondType, _ThirdType]'],
):
    """Interface that allows swapping first and second type values."""

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (Law1(_LawSpec.double_swap_law),)

    @abstractmethod
    def swap(
        self: _SwappableType,
    ) -> KindN[_SwappableType, _SecondType, _FirstType, _ThirdType]:
        """Swaps first and second types in ``SwappableN``."""


#: Type alias for kinds with two type arguments.
Swappable2 = SwappableN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
Swappable3 = SwappableN[_FirstType, _SecondType, _ThirdType]



from copy import copy, deepcopy

import pytest

from returns.primitives.exceptions import ImmutableStateError
from returns.result import Failure, Success


def test_equals():
    """Ensures that ``.equals`` method works correctly."""
    inner_value = 1

    assert Success(inner_value).equals(Success(inner_value))
    assert Failure(inner_value).equals(Failure(inner_value))


def test_not_equals():
    """Ensures that ``.equals`` method works correctly."""
    inner_value = 1

    assert not Success(inner_value).equals(Failure(inner_value))
    assert not Success(inner_value).equals(Success(0))
    assert not Failure(inner_value).equals(Success(inner_value))
    assert not Failure(inner_value).equals(Failure(0))


def test_non_equality():
    """Ensures that containers are not compared to regular values."""
    input_value = 5

    assert Failure(input_value) != input_value
    assert Success(input_value) != input_value
    assert Failure(input_value) != Success(input_value)
    assert hash(Failure(1))
    assert hash(Success(1))


def test_is_compare():
    """Ensures that `is` operator works correctly."""
    left = Failure(1)
    right = Success(1)

    assert left.bind(lambda state: state) is left
    assert right.lash(lambda state: state) is right
    assert right is not Success(1)


def test_immutability_failure():
    """Ensures that Failure container is immutable."""
    with pytest.raises(ImmutableStateError):
        Failure(0)._inner_state = 1  # noqa: SLF001

    with pytest.raises(ImmutableStateError):
        Failure(1).missing = 2

    with pytest.raises(ImmutableStateError):
        del Failure(0)._inner_state  # type: ignore # noqa: SLF001, WPS420

    with pytest.raises(AttributeError):
        Failure(1).missing  # type: ignore # noqa: B018


def test_immutability_success():
    """Ensures that Success container is immutable."""
    with pytest.raises(ImmutableStateError):
        Success(0)._inner_state = 1  # noqa: SLF001

    with pytest.raises(ImmutableStateError):
        Success(1).missing = 2

    with pytest.raises(ImmutableStateError):
        del Success(0)._inner_state  # type: ignore # noqa: SLF001, WPS420

    with pytest.raises(AttributeError):
        Success(1).missing  # type: ignore # noqa: B018


def test_success_immutable_copy():
    """Ensures that Success returns it self when passed to copy function."""
    success = Success(1)
    assert success is copy(success)


def test_success_immutable_deepcopy():
    """Ensures that Success returns it self when passed to deepcopy function."""
    success = Success(1)
    assert success is deepcopy(success)


def test_failure_immutable_copy():
    """Ensures that Failure returns it self when passed to copy function."""
    failure = Failure(0)
    assert failure is copy(failure)


def test_failure_immutable_deepcopy():
    """Ensures that Failure returns it self when passed to deepcopy function."""
    failure = Failure(0)
    assert failure is deepcopy(failure)



from returns.context import RequiresContextIOResultE
from returns.io import IOSuccess


def test_regression394():
    """
    It used to raise ``ImmutableStateError`` for type aliases.

    Here we use the minimal reproduction sample.

    .. code:: python

      Traceback (most recent call last):
        File "ex.py", line 18, in <module>
            get_ip_addr("https://google.com")
        File "ex.py", line 13, in get_ip_addr
            return RequiresContextIOResultE(lambda _: IOSuccess(1))
        File "../3.7.7/lib/python3.7/typing.py", line 677, in __call__
            result.__orig_class__ = self
        File "../returns/returns/primitives/types.py", line 42, in __setattr__
            raise ImmutableStateError()
        returns.primitives.exceptions.ImmutableStateError

    See: https://github.com/dry-python/returns/issues/394

    """
    RequiresContextIOResultE(lambda _: IOSuccess(1))



from typing import ClassVar, final

from mypy.nodes import ARG_OPT, ARG_POS, ARG_STAR, ARG_STAR2, ArgKind
from mypy.typeops import get_type_vars
from mypy.types import (
    AnyType,
    CallableType,
    FunctionLike,
    Overloaded,
    TypeOfAny,
    TypeVarType,
)
from mypy.types import Type as MypyType

from returns.contrib.mypy._structures.args import FuncArg


def proper_type(
    case_functions: list[CallableType],
) -> FunctionLike:
    """Returns a ``CallableType`` or ``Overloaded`` based on case functions."""
    if len(case_functions) == 1:
        return case_functions[0]
    return Overloaded(case_functions)


@final
class Intermediate:
    """
    Allows to build a new callable from old one and different options.

    For example, helps to tell which callee arguments
    was already provided in caller.
    """

    #: Positional arguments can be of this kind.
    _positional_kinds: ClassVar[frozenset[ArgKind]] = frozenset((
        ARG_POS,
        ARG_OPT,
        ARG_STAR,
    ))

    def __init__(self, case_function: CallableType) -> None:
        """We only need a callable to work on."""
        self._case_function = case_function

    def with_applied_args(self, applied_args: list[FuncArg]) -> CallableType:
        """
        By calling this method we construct a new callable from its usage.

        This allows use to create an intermediate callable with just used args.
        """
        new_pos_args = self._applied_positional_args(applied_args)
        new_named_args = self._applied_named_args(applied_args)
        return self.with_signature(new_pos_args + new_named_args)

    def with_signature(self, new_args: list[FuncArg]) -> CallableType:
        """Smartly creates a new callable from a given arguments."""
        return detach_callable(
            self._case_function.copy_modified(
                arg_names=[arg.name for arg in new_args],
                arg_types=[arg.type for arg in new_args],
                arg_kinds=[arg.kind for arg in new_args],
            )
        )

    def with_ret_type(self, ret_type: MypyType) -> CallableType:
        """Smartly creates a new callable from a given return type."""
        return self._case_function.copy_modified(ret_type=ret_type)

    def _applied_positional_args(
        self,
        applied_args: list[FuncArg],
    ) -> list[FuncArg]:
        callee_args = list(
            filter(
                lambda name: (
                    name.name is None
                ),  # TODO: maybe use `kind` instead?
                applied_args,
            )
        )

        new_function_args = []
        for ind, arg in enumerate(FuncArg.from_callable(self._case_function)):
            if arg.kind in self._positional_kinds and ind < len(callee_args):
                new_function_args.append(arg)
        return new_function_args

    def _applied_named_args(
        self,
        applied_args: list[FuncArg],
    ) -> list[FuncArg]:
        callee_args = list(
            filter(
                lambda name: name.name is not None,
                applied_args,
            )
        )

        new_function_args = []
        for arg in FuncArg.from_callable(self._case_function):
            has_named_arg_def = any(
                # Argument can either be used as a named argument
                # or passed to `**kwargs` if it exists.
                arg.name == rdc.name or arg.kind == ARG_STAR2
                for rdc in callee_args
            )
            if callee_args and has_named_arg_def:
                new_function_args.append(arg)
        return new_function_args


@final
class Functions:
    """
    Allows to create new callables based on two existing ones.

    For example, one can need a diff of two callables.
    """

    def __init__(
        self,
        original: CallableType,
        intermediate: CallableType,
    ) -> None:
        """We need two callable to work with."""
        self._original = original
        self._intermediate = intermediate

    def diff(self) -> CallableType:
        """Finds a diff between two functions' arguments."""
        intermediate_names = [
            arg.name for arg in FuncArg.from_callable(self._intermediate)
        ]
        new_function_args = []

        for index, arg in enumerate(FuncArg.from_callable(self._original)):
            should_be_copied = (
                arg.kind in {ARG_STAR, ARG_STAR2}
                or arg.name not in intermediate_names
                or
                # We need to treat unnamed args differently, because python3.8
                # has pos_only_args, all their names are `None`.
                # This is also true for `lambda` functions where `.name`
                # might be missing for some reason.
                (
                    not arg.name
                    and not (
                        index < len(intermediate_names)
                        and
                        # If this is also unnamed arg, then ignoring it.
                        not intermediate_names[index]
                    )
                )
            )
            if should_be_copied:
                new_function_args.append(arg)
        return Intermediate(self._original).with_signature(
            new_function_args,
        )


# TODO: Remove this function once `mypy` order the TypeVars
# by their appearance sequence
def detach_callable(typ: CallableType) -> CallableType:  # noqa: WPS210
    """
    THIS IS A COPY OF `mypy.checker.detach_callable` FUNCTION.

    THE ONLY PURPOSE WE'VE COPIED IS TO GUARANTEE A DETERMINISTIC FOR OUR
    TYPE VARIABLES!
    AS YOU CAN SEE, WE ORDER THE TYPE VARS BY THEIR APPEARANCE SEQUENCE.
    """
    type_list = [*typ.arg_types, typ.ret_type]

    appear_map: dict[str, list[int]] = {}
    for idx, inner_type in enumerate(type_list):
        typevars_available = get_type_vars(inner_type)
        for var in typevars_available:  # noqa: WPS110
            if var.fullname not in appear_map:
                appear_map[var.fullname] = []
            appear_map[var.fullname].append(idx)

    used_type_var_names = appear_map.keys()

    all_type_vars = get_type_vars(typ)
    new_variables = []
    for var in set(all_type_vars):  # noqa: WPS110
        if var.fullname not in used_type_var_names:
            continue
        new_variables.append(
            TypeVarType(
                name=var.name,
                fullname=var.fullname,
                id=var.id,
                values=var.values,
                upper_bound=var.upper_bound,
                variance=var.variance,
                default=AnyType(TypeOfAny.from_omitted_generics),
            ),
        )

    new_variables = sorted(
        new_variables,
        key=lambda item: appear_map[item.fullname][0],  # noqa: WPS110
    )

    return typ.copy_modified(
        variables=new_variables,
        arg_types=type_list[:-1],
        ret_type=type_list[-1],
    )



from collections.abc import Callable
from functools import wraps
from typing import Generic, TypeVar, final

from typing_extensions import ParamSpec

_ReturnType = TypeVar('_ReturnType')
_FuncParams = ParamSpec('_FuncParams')


@final
class Trampoline(Generic[_ReturnType]):
    """
    Represents a wrapped function call.

    Primitive to convert recursion into an actual object.
    """

    __slots__ = ('args', 'func', 'kwargs')

    def __init__(  # noqa: WPS451
        self,
        func: Callable[_FuncParams, _ReturnType],
        /,  # We use pos-only here to be able to store `kwargs` correctly.
        *args: _FuncParams.args,
        **kwargs: _FuncParams.kwargs,
    ) -> None:
        """Save function and given arguments."""
        self.func = getattr(func, '_orig_func', func)
        self.args = args
        self.kwargs = kwargs

    def __call__(self) -> _ReturnType:
        """Call wrapped function with given arguments."""
        return self.func(*self.args, **self.kwargs)


def trampoline(
    func: Callable[_FuncParams, _ReturnType | Trampoline[_ReturnType]],
) -> Callable[_FuncParams, _ReturnType]:
    """
    Convert functions using recursion to regular functions.

    Trampolines allow to unwrap recursion into a regular ``while`` loop,
    which does not raise any ``RecursionError`` ever.

    Since python does not have TCO (tail call optimization),
    we have to provide this helper.

    This is done by wrapping real function calls into
    :class:`returns.trampolines.Trampoline` objects:

    .. code:: python

        >>> from typing import Union
        >>> from returns.trampolines import Trampoline, trampoline

        >>> @trampoline
        ... def get_factorial(
        ...     for_number: int,
        ...     current_number: int = 0,
        ...     acc: int = 1,
        ... ) -> Union[int, Trampoline[int]]:
        ...     assert for_number >= 0
        ...     if for_number <= current_number:
        ...         return acc
        ...     return Trampoline(
        ...         get_factorial,
        ...         for_number,
        ...         current_number=current_number + 1,
        ...         acc=acc * (current_number + 1),
        ...     )

        >>> assert get_factorial(0) == 1
        >>> assert get_factorial(3) == 6
        >>> assert get_factorial(4) == 24

    See also:
        - eli.thegreenplace.net/2017/on-recursion-continuations-and-trampolines
        - https://en.wikipedia.org/wiki/Tail_call

    """

    @wraps(func)
    def decorator(
        *args: _FuncParams.args,
        **kwargs: _FuncParams.kwargs,
    ) -> _ReturnType:
        trampoline_result = func(*args, **kwargs)
        while isinstance(trampoline_result, Trampoline):
            trampoline_result = trampoline_result()
        return trampoline_result

    decorator._orig_func = func  # type: ignore[attr-defined]  # noqa: SLF001
    return decorator



import sys
from collections.abc import Iterable

import pytest

from returns.context import (
    NoDeps,
    Reader,
    ReaderFutureResult,
    ReaderIOResult,
    ReaderResult,
)
from returns.future import Future, FutureFailure, FutureResult, FutureSuccess
from returns.io import IO, IOFailure, IOSuccess
from returns.iterables import Fold
from returns.maybe import Nothing, Some
from returns.result import Failure, Success


def _sum_two(first):
    return lambda second: first + second


@pytest.mark.parametrize(
    ('iterable', 'sequence'),
    [
        # Regular types:
        ([], IO(10)),
        ([IO(1)], IO(11)),
        ([IO(1), IO(2)], IO(13)),
        # Can fail:
        ([], Success(10)),
        ([Success(1)], Success(11)),
        ([Success(1), Success(2)], Success(13)),
        (
            [Failure('a'), Success(1), Success(2)],
            Failure('a'),
        ),
        ([Success(1), Failure('a')], Failure('a')),
        ([Failure('a'), Failure('b')], Failure('a')),
        ([], Some(10)),
        ([Some(1)], Some(11)),
        ([Some(1), Some(2)], Some(13)),
        ([Nothing, Some(1), Some(2)], Nothing),
        ([Some(1), Nothing, Some(2)], Nothing),
        ([Some(1), Some(2), Nothing], Nothing),
        ([Nothing], Nothing),
        ([], IOSuccess(10)),
        ([IOSuccess(1)], IOSuccess(11)),
        ([IOSuccess(1), IOSuccess(2)], IOSuccess(13)),
        (
            [IOFailure('a'), IOSuccess(1), IOSuccess(2)],
            IOFailure('a'),
        ),
        ([IOFailure('a'), IOFailure('b')], IOFailure('a')),
    ],
)
def test_fold_loop(iterable, sequence):
    """Iterable for ``Result`` and ``FailFast``."""
    assert Fold.loop(iterable, sequence.from_value(10), _sum_two) == sequence


@pytest.mark.parametrize(
    ('iterable', 'sequence'),
    [
        # Regular types:
        ([], Reader.from_value(10)),
        ([Reader.from_value(1)], Reader.from_value(11)),
        (
            [Reader.from_value(1), Reader.from_value(2)],
            Reader.from_value(13),
        ),
        # Can fail:
        ([], ReaderResult.from_value(10)),
        ([ReaderResult.from_value(1)], ReaderResult.from_value(11)),
        (
            [ReaderResult.from_value(1), ReaderResult.from_value(2)],
            ReaderResult.from_value(13),
        ),
        (
            [
                ReaderResult.from_failure('a'),
                ReaderResult.from_value(1),
                ReaderResult.from_value(2),
            ],
            ReaderResult.from_failure('a'),
        ),
        (
            [ReaderResult.from_failure('a'), ReaderResult.from_failure('b')],
            ReaderResult.from_failure('a'),
        ),
        ([], ReaderIOResult.from_value(10)),
        ([ReaderIOResult.from_value(1)], ReaderIOResult.from_value(11)),
        (
            [ReaderIOResult.from_value(1), ReaderIOResult.from_value(2)],
            ReaderIOResult.from_value(13),
        ),
        (
            [
                ReaderIOResult.from_failure('a'),
                ReaderIOResult.from_value(1),
                ReaderIOResult.from_value(2),
            ],
            ReaderIOResult.from_failure('a'),
        ),
        (
            [
                ReaderIOResult.from_failure('a'),
                ReaderIOResult.from_failure('b'),
            ],
            ReaderIOResult.from_failure('a'),
        ),
        (
            [ReaderIOResult.from_value(1), ReaderIOResult.from_failure('a')],
            ReaderIOResult.from_failure('a'),
        ),
    ],
)
def test_fold_loop_reader(iterable, sequence):
    """Ensures that ``.loop`` works for readers."""
    assert Fold.loop(
        iterable,
        sequence.from_value(10),
        _sum_two,
    )(...) == sequence(...)


@pytest.mark.anyio
async def test_fold_loop_reader_future_result(subtests):
    """Iterable for ``ReaderFutureResult`` and ``Fold``."""
    containers: list[
        tuple[  # noqa: WPS234
            Iterable[ReaderFutureResult[int, str, NoDeps]],
            ReaderFutureResult[int, str, NoDeps],
        ]
    ] = [
        ([], ReaderFutureResult.from_value(10)),
        (
            [ReaderFutureResult.from_value(1)],
            ReaderFutureResult.from_value(11),
        ),
        (
            [
                ReaderFutureResult.from_value(1),
                ReaderFutureResult.from_value(2),
            ],
            ReaderFutureResult.from_value(13),
        ),
        (
            [
                ReaderFutureResult.from_failure('a'),
                ReaderFutureResult.from_value(1),
                ReaderFutureResult.from_value(2),
            ],
            ReaderFutureResult.from_failure('a'),
        ),
        (
            [
                ReaderFutureResult.from_failure('a'),
                ReaderFutureResult.from_failure('b'),
            ],
            ReaderFutureResult.from_failure('a'),
        ),
        (
            [
                ReaderFutureResult.from_value(1),
                ReaderFutureResult.from_failure('a'),
            ],
            ReaderFutureResult.from_failure('a'),
        ),
    ]
    for iterable, sequence in containers:
        with subtests.test(iterable=iterable, sequence=sequence):
            assert await Fold.loop(
                iterable,
                sequence.from_value(10),
                _sum_two,
            )(...) == await sequence(...)


@pytest.mark.anyio
async def test_fold_collect_future(subtests):
    """Iterable for ``Future`` and ``Fold``."""
    containers: list[
        tuple[  # noqa: WPS234
            Iterable[Future[int]],
            Future[int],
        ]
    ] = [
        ([], Future.from_value(10)),
        ([Future.from_value(1)], Future.from_value(11)),
        (
            [Future.from_value(1), Future.from_value(2)],
            Future.from_value(13),
        ),
    ]
    for iterable, sequence in containers:
        with subtests.test(iterable=iterable, sequence=sequence):
            assert (
                await Fold.loop(
                    iterable,
                    sequence.from_value(10),
                    _sum_two,
                )
                == await sequence
            )


@pytest.mark.anyio
async def test_fold_collect_future_result(subtests):
    """Iterable for ``FutureResult`` and ``Fold``."""
    containers: list[
        tuple[  # noqa: WPS234
            Iterable[FutureResult[int, str]],
            FutureResult[int, str],
        ]
    ] = [
        ([], FutureSuccess(10)),
        ([FutureSuccess(1)], FutureSuccess(11)),
        ([FutureSuccess(1), FutureSuccess(2)], FutureSuccess(13)),
        (
            [FutureFailure('a'), FutureSuccess(1), FutureSuccess(2)],
            FutureFailure('a'),
        ),
        ([FutureFailure('a'), FutureFailure('b')], FutureFailure('a')),
        ([FutureSuccess(1), FutureFailure('a')], FutureFailure('a')),
    ]
    for iterable, sequence in containers:
        with subtests.test(iterable=iterable, sequence=sequence):
            assert (
                await Fold.loop(
                    iterable,
                    sequence.from_value(10),
                    _sum_two,
                )
                == await sequence
            )


def test_fold_loop_recursion_limit():
    """Ensures that ``.loop`` method is recursion safe."""
    limit = sys.getrecursionlimit() + 1
    iterable = (IO(1) for _ in range(limit))
    assert Fold.loop(iterable, IO(0), _sum_two) == IO(limit)



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, ClassVar, TypeVar, final

from returns.interfaces.specific import ioresult, reader, reader_result
from returns.primitives.hkt import KindN
from returns.primitives.laws import (
    Law,
    Law2,
    Lawful,
    LawSpecDef,
    law_definition,
)

if TYPE_CHECKING:
    from returns.context import ReaderIOResult  # noqa: WPS433
    from returns.io import IOResult  # noqa: F401, WPS433

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_ValueType = TypeVar('_ValueType')
_ErrorType = TypeVar('_ErrorType')
_EnvType = TypeVar('_EnvType')

_ReaderIOResultLikeType = TypeVar(
    '_ReaderIOResultLikeType',
    bound='ReaderIOResultLikeN',
)


class ReaderIOResultLikeN(
    reader_result.ReaderResultLikeN[_FirstType, _SecondType, _ThirdType],
    ioresult.IOResultLikeN[_FirstType, _SecondType, _ThirdType],
):
    """
    Base interface for all types that do look like ``ReaderIOResult`` instance.

    Cannot be called.
    """

    __slots__ = ()

    @abstractmethod
    def bind_context_ioresult(
        self: _ReaderIOResultLikeType,
        function: Callable[
            [_FirstType],
            ReaderIOResult[_UpdatedType, _SecondType, _ThirdType],
        ],
    ) -> KindN[_ReaderIOResultLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Binds a ``ReaderIOResult`` returning function over a container."""

    @classmethod
    @abstractmethod
    def from_ioresult_context(
        cls: type[_ReaderIOResultLikeType],
        inner_value: ReaderIOResult[_ValueType, _ErrorType, _EnvType],
    ) -> KindN[_ReaderIOResultLikeType, _ValueType, _ErrorType, _EnvType]:
        """Unit method to create new containers from ``ReaderIOResult``."""


#: Type alias for kinds with three type arguments.
ReaderIOResultLike3 = ReaderIOResultLikeN[_FirstType, _SecondType, _ThirdType]


@final
class _LawSpec(LawSpecDef):
    """
    Concrete laws for ``ReaderIOResultBasedN``.

    See: https://github.com/haskell/mtl/pull/61/files
    """

    __slots__ = ()

    @law_definition
    def asking_law(
        container: ReaderIOResultBasedN[_FirstType, _SecondType, _ThirdType],
        env: _ThirdType,
    ) -> None:
        """Asking for an env, always returns the env."""
        assert container.ask().__call__(  # noqa: PLC2801
            env,
        ) == container.from_value(env).__call__(env)  # noqa: PLC2801


class ReaderIOResultBasedN(
    ReaderIOResultLikeN[_FirstType, _SecondType, _ThirdType],
    reader.CallableReader3[
        _FirstType,
        _SecondType,
        _ThirdType,
        # Calls:
        'IOResult[_FirstType, _SecondType]',
        _ThirdType,
    ],
    Lawful['ReaderIOResultBasedN[_FirstType, _SecondType, _ThirdType]'],
):
    """
    This interface is very specific to our ``ReaderIOResult`` type.

    The only thing that differs from ``ReaderIOResultLikeN`` is that we know
    the specific types for its ``__call__`` method.

    In this case the return type of ``__call__`` is ``IOResult``.
    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (Law2(_LawSpec.asking_law),)


#: Type alias for kinds with three type arguments.
ReaderIOResultBased3 = ReaderIOResultBasedN[_FirstType, _SecondType, _ThirdType]



from returns.io import IOFailure, IOSuccess


def test_equals():
    """Ensures that ``.equals`` method works correctly."""
    inner_value = 1

    assert IOSuccess(inner_value).equals(IOSuccess(inner_value))
    assert IOFailure(inner_value).equals(IOFailure(inner_value))


def test_not_equals():
    """Ensures that ``.equals`` method works correctly."""
    inner_value = 1

    assert not IOSuccess(inner_value).equals(IOFailure(inner_value))
    assert not IOSuccess(inner_value).equals(IOSuccess(0))
    assert not IOFailure(inner_value).equals(IOSuccess(inner_value))
    assert not IOFailure(inner_value).equals(IOFailure(0))



import inspect
import sys
from collections.abc import Callable, Iterator
from contextlib import ExitStack, contextmanager
from functools import partial, wraps
from types import FrameType, MappingProxyType
from typing import TYPE_CHECKING, Any, Final, TypeAlias, TypeVar, Union, final
from unittest import mock

import pytest

if TYPE_CHECKING:
    from returns.interfaces.specific.result import ResultLikeN

# We keep track of errors handled by keeping a mapping of <object id>: object.
# If an error is handled, it is in the mapping.
# If it isn't in the mapping, the error is not handled.
#
# Note only storing object IDs would not work, as objects may be GC'ed
# and their object id assigned to another object.
# Also, the object itself cannot be (in) the key because
# (1) we cannot always assume hashability and
# (2) we need to track the object identity, not its value
_ErrorsHandled: TypeAlias = dict[int, Any]

_FunctionType = TypeVar('_FunctionType', bound=Callable)
_ReturnsResultType = TypeVar(
    '_ReturnsResultType',
    bound=Union['ResultLikeN', Callable[..., 'ResultLikeN']],
)


@final
class ReturnsAsserts:
    """Class with helpers assertions to check containers."""

    __slots__ = ('_errors_handled',)

    def __init__(self, errors_handled: _ErrorsHandled) -> None:
        """Constructor for this type."""
        self._errors_handled = errors_handled

    @staticmethod  # noqa: WPS602
    def assert_equal(  # noqa: WPS602
        first,
        second,
        *,
        deps=None,
        backend: str = 'asyncio',
    ) -> None:
        """Can compare two containers even with extra calling and awaiting."""
        from returns.primitives.asserts import assert_equal  # noqa: PLC0415

        assert_equal(first, second, deps=deps, backend=backend)

    def is_error_handled(self, container) -> bool:
        """Ensures that container has its error handled in the end."""
        return id(container) in self._errors_handled

    @staticmethod  # noqa: WPS602
    @contextmanager
    def assert_trace(  # noqa: WPS602
        trace_type: _ReturnsResultType,
        function_to_search: _FunctionType,
    ) -> Iterator[None]:
        """
        Ensures that a given function was called during execution.

        Use it to determine where the failure happened.
        """
        old_tracer = sys.gettrace()
        sys.settrace(partial(_trace_function, trace_type, function_to_search))

        try:
            yield
        except _DesiredFunctionFound:
            pass  # noqa: WPS420
        else:
            pytest.fail(
                f'No container {type(trace_type).__name__} was created',
            )
        finally:
            sys.settrace(old_tracer)


def _trace_function(
    trace_type: _ReturnsResultType,
    function_to_search: _FunctionType,
    frame: FrameType,
    event: str,
    arg: Any,
) -> None:
    is_desired_type_call = event == 'call' and (
        # Some containers is created through functions and others
        # is created directly using class constructors!
        # The first line covers when it's created through a function
        # The second line covers when it's created through a
        # class constructor
        frame.f_code is getattr(trace_type, '__code__', None)
        or frame.f_code is getattr(trace_type.__init__, '__code__', None)  # type: ignore[misc]
    )
    if is_desired_type_call:
        current_call_stack = inspect.stack()
        function_to_search_code = getattr(function_to_search, '__code__', None)
        for frame_info in current_call_stack:
            if function_to_search_code is frame_info.frame.f_code:
                raise _DesiredFunctionFound


class _DesiredFunctionFound(BaseException):  # noqa: WPS418
    """Exception to raise when expected function is found."""


def pytest_configure(config) -> None:
    """
    Hook to be executed on import.

    We use it define custom markers.
    """
    config.addinivalue_line(
        'markers',
        (
            'returns_lawful: all tests under `check_all_laws` '
            + 'is marked this way, '
            + 'use `-m "not returns_lawful"` to skip them.'
        ),
    )


@pytest.fixture
def returns() -> Iterator[ReturnsAsserts]:
    """Returns class with helpers assertions to check containers."""
    with _spy_error_handling() as errors_handled:
        yield ReturnsAsserts(errors_handled)


@contextmanager
def _spy_error_handling() -> Iterator[_ErrorsHandled]:
    """Track error handling of containers."""
    errs: _ErrorsHandled = {}
    with ExitStack() as cleanup:
        for container in _containers_to_patch():
            for method, patch in _ERROR_HANDLING_PATCHERS.items():
                cleanup.enter_context(
                    mock.patch.object(
                        container,
                        method,
                        patch(getattr(container, method), errs=errs),
                    )
                )
        yield errs


# delayed imports are needed to prevent messing up coverage
def _containers_to_patch() -> list:
    from returns.context import (  # noqa: PLC0415
        RequiresContextFutureResult,
        RequiresContextIOResult,
        RequiresContextResult,
    )
    from returns.future import FutureResult  # noqa: PLC0415
    from returns.io import IOFailure, IOSuccess  # noqa: PLC0415
    from returns.result import Failure, Success  # noqa: PLC0415

    return [
        Success,
        Failure,
        IOSuccess,
        IOFailure,
        RequiresContextResult,
        RequiresContextIOResult,
        RequiresContextFutureResult,
        FutureResult,
    ]


def _patched_error_handler(
    original: _FunctionType,
    errs: _ErrorsHandled,
) -> _FunctionType:
    if inspect.iscoroutinefunction(original):

        async def wrapper(self, *args, **kwargs):
            original_result = await original(self, *args, **kwargs)
            errs[id(original_result)] = original_result
            return original_result

    else:

        def wrapper(self, *args, **kwargs):
            original_result = original(self, *args, **kwargs)
            errs[id(original_result)] = original_result
            return original_result

    return wraps(original)(wrapper)  # type: ignore


def _patched_error_copier(
    original: _FunctionType,
    errs: _ErrorsHandled,
) -> _FunctionType:
    if inspect.iscoroutinefunction(original):

        async def wrapper(self, *args, **kwargs):
            original_result = await original(self, *args, **kwargs)
            if id(self) in errs:
                errs[id(original_result)] = original_result
            return original_result

    else:

        def wrapper(self, *args, **kwargs):
            original_result = original(self, *args, **kwargs)
            if id(self) in errs:
                errs[id(original_result)] = original_result
            return original_result

    return wraps(original)(wrapper)  # type: ignore


_ERROR_HANDLING_PATCHERS: Final = MappingProxyType({
    'lash': _patched_error_handler,
    'map': _patched_error_copier,
    'alt': _patched_error_copier,
})



from returns.context import (
    ReaderIOResult,
    ReaderIOResultE,
    RequiresContextIOResult,
    RequiresContextIOResultE,
)


def _function(arg: int) -> RequiresContextIOResultE[float, int]:
    if arg == 0:
        return RequiresContextIOResult.from_failure(
            ZeroDivisionError('Divided by 0'),
        )
    return RequiresContextIOResult.from_value(10 / arg)


def test_requires_context_ioresulte():
    """Ensures that RequiresContextIOResultE correctly typecast."""
    container: RequiresContextIOResult[float, Exception, int] = _function(1)
    assert container(0) == RequiresContextIOResult.from_value(10.0)(0)


def test_requires_context_io_aliases():
    """Ensures that ReaderIOResult correctly typecast."""
    container: ReaderIOResultE[float, int] = _function(1)
    container2: ReaderIOResult[float, Exception, int] = _function(1)
    container3: ReaderIOResultE[float, int] = ReaderIOResultE.from_value(
        10.0,
    )
    container4: ReaderIOResultE[float, int] = ReaderIOResult.from_value(10.0)

    assert container(0) == container2(0) == container3(0) == container4(0)
    assert container(0) == RequiresContextIOResult.from_value(10.0)(0)



from collections.abc import Callable

from returns.context import RequiresContext


def _same_function(some_arg: int) -> Callable[[float], float]:
    return lambda other: other / some_arg


def test_equality() -> None:
    """Ensures that containers can be compared."""
    assert RequiresContext(_same_function) == RequiresContext(_same_function)


def test_nonequality() -> None:
    """Ensures that containers can be compared."""
    assert RequiresContext(_same_function) != RequiresContext(str)
    assert RequiresContext.from_value(1) != RequiresContext.from_value(1)



from returns.io import IO
from returns.unsafe import unsafe_perform_io


def test_unsafe_perform_io():
    """Ensures that unsafe_perform_io returns the object itself."""
    id_object = object()
    assert unsafe_perform_io(IO(id_object)) is id_object



from typing import TypeVar, overload

from returns.functions import identity
from returns.interfaces.bindable import BindableN
from returns.maybe import Maybe, Nothing, Some
from returns.pipeline import is_successful
from returns.primitives.hkt import KindN, kinded
from returns.result import Failure, Result, Success

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')

_BindableKind = TypeVar('_BindableKind', bound=BindableN)


@kinded
def flatten(
    container: KindN[
        _BindableKind,
        KindN[_BindableKind, _FirstType, _SecondType, _ThirdType],
        _SecondType,
        _ThirdType,
    ],
) -> KindN[_BindableKind, _FirstType, _SecondType, _ThirdType]:
    """
    Joins two nested containers together.

    Please, note that it will not join
    two ``Failure`` for ``Result`` case
    or two ``Nothing`` for ``Maybe`` case
    (or basically any two error types) together.

    .. code:: python

      >>> from returns.converters import flatten
      >>> from returns.io import IO
      >>> from returns.result import Failure, Success

      >>> assert flatten(IO(IO(1))) == IO(1)

      >>> assert flatten(Success(Success(1))) == Success(1)
      >>> assert flatten(Failure(Failure(1))) == Failure(Failure(1))

    See also:
        - https://bit.ly/2sIviUr

    """
    return container.bind(identity)


def result_to_maybe(
    result_container: Result[_FirstType, _SecondType],
) -> Maybe[_FirstType]:
    """
    Converts ``Result`` container to ``Maybe`` container.

    .. code:: python

      >>> from returns.maybe import Some, Nothing
      >>> from returns.result import Failure, Success

      >>> assert result_to_maybe(Success(1)) == Some(1)
      >>> assert result_to_maybe(Success(None)) == Some(None)
      >>> assert result_to_maybe(Failure(1)) == Nothing
      >>> assert result_to_maybe(Failure(None)) == Nothing

    """
    if is_successful(result_container):
        return Some(result_container.unwrap())
    return Nothing


@overload
def maybe_to_result(
    maybe_container: Maybe[_FirstType],
) -> Result[_FirstType, None]: ...


@overload
def maybe_to_result(
    maybe_container: Maybe[_FirstType],
    default_error: _SecondType,
) -> Result[_FirstType, _SecondType]: ...


def maybe_to_result(
    maybe_container: Maybe[_FirstType],
    default_error: _SecondType | None = None,
) -> Result[_FirstType, _SecondType | None]:
    """
    Converts ``Maybe`` container to ``Result`` container.

    With optional ``default_error`` to be used for ``Failure``'s error value.

    .. code:: python

      >>> from returns.maybe import Some, Nothing
      >>> from returns.result import Failure, Success

      >>> assert maybe_to_result(Some(1)) == Success(1)
      >>> assert maybe_to_result(Some(None)) == Success(None)
      >>> assert maybe_to_result(Nothing) == Failure(None)

      >>> assert maybe_to_result(Nothing, 'error') == Failure('error')

    """
    if is_successful(maybe_container):
        return Success(maybe_container.unwrap())
    return Failure(default_error)



from collections.abc import Callable
from functools import wraps
from typing import TypeVar

from mypy.types import AnyType, TypeOfAny

_CallableType = TypeVar('_CallableType', bound=Callable)


def asserts_fallback_to_any(function: _CallableType) -> _CallableType:
    """
    Falls back to ``Any`` when some ``assert ...`` fails in our plugin code.

    We often use ``assert isinstance(variable, Instance)``
    as a way to ensure correctness in this plugin.
    But, we need a generic way to handle all
    possible exceptions in a single manner:
    just return ``Any`` and hope that someday someone reports it.

    """

    @wraps(function)
    def decorator(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except AssertionError:
            # TODO: log it somehow
            return AnyType(TypeOfAny.implementation_artifact)

    return decorator  # type: ignore



from returns.pointfree.alt import alt as alt
from returns.pointfree.apply import apply as apply
from returns.pointfree.bimap import bimap as bimap
from returns.pointfree.bind import bind as bind
from returns.pointfree.bind_async import bind_async as bind_async
from returns.pointfree.bind_async_context_future_result import (
    bind_async_context_future_result as bind_async_context_future_result,
)
from returns.pointfree.bind_async_future import (
    bind_async_future as bind_async_future,
)
from returns.pointfree.bind_async_future_result import (
    bind_async_future_result as bind_async_future_result,
)
from returns.pointfree.bind_awaitable import bind_awaitable as bind_awaitable
from returns.pointfree.bind_context import bind_context as bind_context
from returns.pointfree.bind_context import bind_context2 as bind_context2
from returns.pointfree.bind_context import bind_context3 as bind_context3
from returns.pointfree.bind_context_future_result import (
    bind_context_future_result as bind_context_future_result,
)
from returns.pointfree.bind_context_ioresult import (
    bind_context_ioresult as bind_context_ioresult,
)
from returns.pointfree.bind_context_result import (
    bind_context_result as bind_context_result,
)
from returns.pointfree.bind_future import bind_future as bind_future
from returns.pointfree.bind_future_result import (
    bind_future_result as bind_future_result,
)
from returns.pointfree.bind_io import bind_io as bind_io
from returns.pointfree.bind_ioresult import bind_ioresult as bind_ioresult
from returns.pointfree.bind_optional import bind_optional as bind_optional
from returns.pointfree.bind_result import bind_result as bind_result
from returns.pointfree.compose_result import compose_result as compose_result
from returns.pointfree.cond import cond as cond
from returns.pointfree.lash import lash as lash
from returns.pointfree.map import map_ as map_
from returns.pointfree.modify_env import modify_env as modify_env
from returns.pointfree.modify_env import modify_env2 as modify_env2
from returns.pointfree.modify_env import modify_env3 as modify_env3
from returns.pointfree.unify import unify as unify



from typing import TypeVar, overload

from returns.context import NoDeps
from returns.interfaces.failable import DiverseFailableN, SingleFailableN
from returns.primitives.hkt import KindN, kinded

_ValueType = TypeVar('_ValueType')
_ErrorType = TypeVar('_ErrorType')

_SingleFailableKind = TypeVar('_SingleFailableKind', bound=SingleFailableN)
_DiverseFailableKind = TypeVar('_DiverseFailableKind', bound=DiverseFailableN)


@overload
def internal_cond(
    container_type: type[_SingleFailableKind],
    is_success: bool,  # noqa: FBT001
    success_value: _ValueType,
) -> KindN[_SingleFailableKind, _ValueType, _ErrorType, NoDeps]: ...


@overload
def internal_cond(
    container_type: type[_DiverseFailableKind],
    is_success: bool,  # noqa: FBT001
    success_value: _ValueType,
    error_value: _ErrorType,
) -> KindN[_DiverseFailableKind, _ValueType, _ErrorType, NoDeps]: ...


def internal_cond(
    container_type: (type[_SingleFailableKind] | type[_DiverseFailableKind]),
    is_success: bool,  # noqa: FBT001
    success_value: _ValueType,
    error_value: _ErrorType | None = None,
):
    """
    Reduce the boilerplate when choosing paths.

    Works with ``SingleFailableN`` (e.g. ``Maybe``)
    and ``DiverseFailableN`` (e.g. ``Result``).

    Example using ``cond`` with the ``Result`` container:

    .. code:: python

      >>> from returns.methods import cond
      >>> from returns.result import Failure, Result, Success

      >>> def is_numeric(string: str) -> Result[str, str]:
      ...     return cond(
      ...         Result,
      ...         string.isnumeric(),
      ...         'It is a number',
      ...         'It is not a number',
      ...     )

      >>> assert is_numeric('42') == Success('It is a number')
      >>> assert is_numeric('non numeric') == Failure('It is not a number')

    Example using ``cond`` with the ``Maybe`` container:

    .. code:: python

      >>> from returns.maybe import Maybe, Some, Nothing

      >>> def is_positive(number: int) -> Maybe[int]:
      ...     return cond(Maybe, number > 0, number)

      >>> assert is_positive(10) == Some(10)
      >>> assert is_positive(-10) == Nothing

    """
    if is_success:
        return container_type.from_value(success_value)

    if issubclass(container_type, DiverseFailableN):
        return container_type.from_failure(error_value)
    return container_type.empty


#: Kinded version of :func:`~internal_cond`, use it to infer real return type.
cond = kinded(internal_cond)
